import numpy as np
from .. import accelerometer as accel


class TestActivationFeatures:
    def test_activation(self):
        # test on single sample multi-channel signal
        X = np.array([[0, 1, 0]])
        result = accel.activation_features(X)
        np.testing.assert_array_equal(result[0], np.array(
            [[0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0]]))

        # test on single sample single-channel signal
        X = np.array([[0, ]])
        result = accel.activation_features(X)
        np.testing.assert_array_equal(result[0], np.array(
            [[0, 0, 0, 0]]))
        X = np.array([[1, ]])
        result = accel.activation_features(X)
        np.testing.assert_array_equal(result[0], np.array(
            [[1, 1, 1, 0]]))

        # test on multi sample multi-channel signal edge case
        X = np.concatenate(
            (np.zeros((5, 3)), np.ones((5, 3)), np.zeros((5, 3))), axis=0)
        result = accel.activation_features(X)
        np.testing.assert_array_equal(result[0], np.array(
            [[5/15, 5/15, 5/15, 1/5, 1/5, 1/5, 5/15, 5/15, 5/15, 0, 0, 0]]))

        X = np.concatenate(
            (np.zeros((5, 3)), np.ones((5, 3))), axis=0)
        result = accel.activation_features(X)
        np.testing.assert_array_equal(result[0], np.array(
            [[5/10, 5/10, 5/10, 1/5, 1/5, 1/5, 5/10, 5/10, 5/10, 0, 0, 0]]))

        # test on multi sample multi-channel signal edge case
        X = np.concatenate(
            (np.ones((5, 3)), np.zeros((5, 3))), axis=0)
        result = accel.activation_features(X)
        np.testing.assert_array_equal(result[0], np.array(
            [[5/10, 5/10, 5/10, 1/5, 1/5, 1/5, 5/10, 5/10, 5/10, 0, 0, 0]]))

        # test on multi sample multi-channel signal edge case
        X = np.ones((10, 3))
        result = accel.activation_features(X)
        np.testing.assert_array_equal(result[0], np.array(
            [[10/10, 10/10, 10/10, 1/10, 1/10, 1/10, 10/10, 10/10, 10/10, 0, 0, 0]]))

        X = np.zeros((10, 3))
        result = accel.activation_features(X)
        np.testing.assert_array_equal(result[0], np.array(
            [[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]]))

        # test multiple activations
        X = np.concatenate(
            (np.zeros((5, 3)), np.ones((5, 3)), np.zeros((5, 3)), np.ones((10, 3))), axis=0)
        result = accel.activation_features(X)
        np.testing.assert_array_almost_equal(result[0], np.array(
            [[15/25, 15/25, 15/25, 2/15, 2/15, 2/15, 7.5/25, 7.5/25, 7.5/25, 3.53553/25, 3.53553/25, 3.53553/25]]))


class TestCounts:
    def test_enmo(self):
        # test on a single row
        X = np.array([[1., 1., 1., ]])
        result = accel.enmo(X)
        assert np.allclose(result[0], np.sqrt(3) - 1, atol=0.001)
        assert result[1] == 'ENMO_0'

        # test on a single row with nan
        X = np.array([[1., np.nan, 1., ]])
        result = accel.enmo(X)
        assert np.allclose(result[0], np.nan, atol=0.001, equal_nan=True)
        assert result[1] == 'ENMO_0'

        # test on an array
        X = np.array([[1., 1., 1., ], [1., 1., 1.]])
        result = accel.enmo(X)
        np.testing.assert_array_equal(result[0], np.array(
            [[np.sqrt(3) - 1]]))
        assert result[1] == 'ENMO_0'

        # test on an array with nan
        X = np.array([[1., 1., 1., ], [2., np.nan, 2.]])
        result = accel.enmo(X)
        np.testing.assert_array_equal(result[0], np.array([[np.sqrt(3) - 1]]))
        assert result[1] == 'ENMO_0'


class TestOrientation:
    def test_gravity_angles(self):
        # test without subwins
        # test angles (0, 90, 90) with unit in degree on 1d array
        X = np.array([[1., 0., 0.]])
        result = accel.gravity_angles(X, unit='deg')
        np.testing.assert_array_equal(result[0], np.array([[0, 90, 90]]))
        # test angles (90, 0, 90) with unit in degree on 1d array
        X = np.array([[0, 1., 0.]])
        result = accel.gravity_angles(X, unit='deg')
        np.testing.assert_array_equal(result[0], np.array([[90, 0, 90]]))
        # test angles (90, 90, 0) with unit in degree on 1d array
        X = np.array([[0, 0., 1.]])
        result = accel.gravity_angles(X, unit='deg')
        np.testing.assert_array_equal(result[0], np.array([[90, 90, 0]]))
        # test angles (45, 45, 90) with unit in degree on 1d array
        X = np.array([[np.sqrt(2)/2, np.sqrt(2)/2, 0.]])
        result = accel.gravity_angles(X, unit='deg')
        np.testing.assert_array_equal(
            result[0], np.array([[45, 45, 90]]))

        # test angles (0, 90, 90) with unit in radian on 1d array
        X = np.array([[1., 0., 0.]])
        result = accel.gravity_angles(X, unit='rad')
        np.testing.assert_array_equal(result[0], np.array(
            [[np.deg2rad(0), np.deg2rad(90), np.deg2rad(90)]]))

        # test angles (0, 90, 90) with unit in degree on 2d array
        X = np.array([[1., 0., 0.], [1., 0., 0.]])
        result = accel.gravity_angles(X, unit='deg')
        np.testing.assert_array_equal(result[0], np.array(
            [[0, 90, 90]]))

        # test angles (0, 90, 90) with unit in degree on 2d array containing nan
        X = np.array([[1., 0., 0.], [1., np.nan, 0.]])
        result = accel.gravity_angles(X, unit='deg')
        np.testing.assert_array_equal(result[0], np.array(
            [[0, 90, 90]]))

        # test angles (0, 90, 90) with unit in degree on 2d array containing nan in the whole column
        X = np.array([[1., np.nan, 0.], [1., np.nan, 0.]])
        result = accel.gravity_angles(X, unit='deg')
        np.testing.assert_array_equal(result[0], np.array([[np.nan] * 3]))

        # test on a single row with nan
        X = np.array([[1., np.nan, 0., ]])
        result = accel.gravity_angles(X, unit='deg')
        np.testing.assert_array_equal(result[0], np.array(
            [[np.nan, np.nan, np.nan]]))
        np.testing.assert_array_equal(
            result[1], ["G_ANGLE_X_0", "G_ANGLE_Y_0", "G_ANGLE_Z_0"])

        # test on a 2d array with subwins
        X = np.concatenate([np.tile([[0., -1., 0., ]], (5, 1)),
                            np.tile([[1., 0., 0., ]], (5, 1)),
                            np.tile([[0., 0., 1., ]], (5, 1))], axis=0)
        result = accel.gravity_angles(X, subwins=3, unit='deg')
        np.testing.assert_array_almost_equal(result[0], np.array(
            [[90, 180, 90, 0, 90, 90, 90, 90, 0]]))
        np.testing.assert_array_equal(result[1], ["G_ANGLE_X_0", "G_ANGLE_Y_0", "G_ANGLE_Z_0", "G_ANGLE_X_1",
                                                  "G_ANGLE_Y_1", "G_ANGLE_Z_1", "G_ANGLE_X_2", "G_ANGLE_Y_2", "G_ANGLE_Z_2"])

    def test_orientation_features(self):
        # test on a single row
        X = np.array([[1., 0., 0., ]])
        result = accel.orientation_features(X, unit='deg')
        np.testing.assert_array_equal(result[0], np.array(
            [[0, 90, 90, 0, 0, 0, np.nan, np.nan, np.nan]]))
        np.testing.assert_array_equal(result[1], ["MEDIAN_G_ANGLE_0",
                                                  "MEDIAN_G_ANGLE_1", "MEDIAN_G_ANGLE_2", "RANGE_G_ANGLE_0",
                                                  "RANGE_G_ANGLE_1", "RANGE_G_ANGLE_2", "STD_G_ANGLE_0", "STD_G_ANGLE_1", "STD_G_ANGLE_2"])

        # test on a single row with nan
        X = np.array([[1., np.nan, 0., ]])
        result = accel.orientation_features(X, unit='deg')
        np.testing.assert_array_equal(result[0], np.full((1, 9), np.nan))
        np.testing.assert_array_equal(result[1], ["MEDIAN_G_ANGLE_0",
                                                  "MEDIAN_G_ANGLE_1", "MEDIAN_G_ANGLE_2", "RANGE_G_ANGLE_0",
                                                  "RANGE_G_ANGLE_1", "RANGE_G_ANGLE_2", "STD_G_ANGLE_0", "STD_G_ANGLE_1", "STD_G_ANGLE_2"])

        # test on a 2d array with subwins
        X = np.concatenate([np.tile([[0., -1., 0., ]], (5, 1)),
                            np.tile([[1., 0., 0., ]], (5, 1)),
                            np.tile([[0., 0., 1., ]], (5, 1))], axis=0)
        result = accel.orientation_features(X, subwins=3, unit='deg')
        np.testing.assert_array_almost_equal(result[0], np.array(
            [[90, 90, 90, 90, 90, 90, 51.961524, 51.961524, 51.961524]]))
        np.testing.assert_array_equal(result[1], ["MEDIAN_G_ANGLE_0",
                                                  "MEDIAN_G_ANGLE_1", "MEDIAN_G_ANGLE_2", "RANGE_G_ANGLE_0",
                                                  "RANGE_G_ANGLE_1", "RANGE_G_ANGLE_2", "STD_G_ANGLE_0", "STD_G_ANGLE_1", "STD_G_ANGLE_2"])


class TestSpectrum:
    def test_spectrum_features(self):
        names = ['DOM_FREQ_0',
                 'DOM_FREQ_POWER_0',
                 'TOTAL_FREQ_POWER_0',
                 'FREQ_POWER_ABOVE_3DOT5_0',
                 'FREQ_POWER_RATIO_ABOVE_3DOT5_0',
                 'DOM_FREQ_POWER_RATIO_0',
                 'FREQ_POWER_RATIO_BEWTEEN_DOT5_2DOT5_0',
                 'DOM_FREQ_POWER_BETWEEN_DOT6_2DOT6_0',
                 'DOM_FREQ_RATIO_PREV_BOUT_0',
                 'SPECTRAL_ENTROPY_0']
        # test on single sample single channel signal
        X = np.array([[0, ]])
        result = accel.spectrum_features(
            X, sr=80, freq_range=None, prev_spectrum_features=None)
        np.testing.assert_array_equal(
            result[0], np.array(
                [[0, 0, 0, 0, 0, 0, 0, np.nan, np.nan, np.nan]]))
        np.testing.assert_array_equal(result[1], names)

        # test on multiple sample single channel signal
        dom_freq = 1
        sr = 100
        X = np.atleast_2d(np.sin(2*np.pi * dom_freq *
                                 np.arange(0, 1, 1.0 / sr))).T
        result = accel.spectrum_features(
            X, sr=100, freq_range=None, prev_spectrum_features=None)
        np.testing.assert_array_almost_equal(
            result[0], np.array(
                [[1, 4.283017e-01, 6.107265e-01, 0, 0, 4.283017e-01 / 6.107265e-01, 1, 4.283017e-01, np.nan, 0.15508315]]))
        np.testing.assert_array_equal(result[1], names)

        # test on multiple sample single channel signal with nan
        dom_freq = 1
        sr = 100
        X = np.atleast_2d(np.sin(2*np.pi * dom_freq *
                                 np.arange(0, 1, 1.0 / sr))).T
        X[5:10, 0] = np.nan
        result = accel.spectrum_features(
            X, sr=100, freq_range=None, prev_spectrum_features=None)
        np.testing.assert_array_almost_equal(
            result[0], np.array(
                [[1, 4.283017e-01, 6.107265e-01, 0, 0, 4.283017e-01 / 6.107265e-01, 1, 4.283017e-01, np.nan, 0.15508315]]), decimal=4)
        np.testing.assert_array_equal(result[1], names)

        # test on single sample multi channel signal
        names = ['DOM_FREQ_0', 'DOM_FREQ_1', 'DOM_FREQ_2'] + \
            ['DOM_FREQ_POWER_0', 'DOM_FREQ_POWER_1', 'DOM_FREQ_POWER_2'] + \
            ['TOTAL_FREQ_POWER_0', 'TOTAL_FREQ_POWER_1', 'TOTAL_FREQ_POWER_2'] + \
            ['FREQ_POWER_ABOVE_3DOT5_0', 'FREQ_POWER_ABOVE_3DOT5_1', 'FREQ_POWER_ABOVE_3DOT5_2'] + \
            ['FREQ_POWER_RATIO_ABOVE_3DOT5_0', 'FREQ_POWER_RATIO_ABOVE_3DOT5_1', 'FREQ_POWER_RATIO_ABOVE_3DOT5_2'] + \
            ['DOM_FREQ_POWER_RATIO_0', 'DOM_FREQ_POWER_RATIO_1', 'DOM_FREQ_POWER_RATIO_2'] + \
            ['FREQ_POWER_RATIO_BEWTEEN_DOT5_2DOT5_0', 'FREQ_POWER_RATIO_BEWTEEN_DOT5_2DOT5_1', 'FREQ_POWER_RATIO_BEWTEEN_DOT5_2DOT5_2'] + \
            ['DOM_FREQ_POWER_BETWEEN_DOT6_2DOT6_0', 'DOM_FREQ_POWER_BETWEEN_DOT6_2DOT6_1', 'DOM_FREQ_POWER_BETWEEN_DOT6_2DOT6_2'] + \
            ['DOM_FREQ_RATIO_PREV_BOUT_0', 'DOM_FREQ_RATIO_PREV_BOUT_1',
                'DOM_FREQ_RATIO_PREV_BOUT_2', 'SPECTRAL_ENTROPY_0', 'SPECTRAL_ENTROPY_1', 'SPECTRAL_ENTROPY_2']
        sr = 100
        X = np.array([[0, 0, 0]])
        result = accel.spectrum_features(
            X, sr=100, freq_range=None, prev_spectrum_features=None)
        np.testing.assert_array_almost_equal(
            result[0], np.array(
                [[0] * 3 +
                 [0] * 3 +
                 [0] * 3 +
                 [0] * 3 +
                 [0] * 3 +
                 [0] * 3 +
                 [0] * 3 +
                 [np.nan] * 3 +
                 [np.nan] * 3 +
                 [np.nan] * 3]))
        np.testing.assert_array_equal(result[1], names)

        # test on multiple sample multi channel signal
        dom_freq = 1
        sr = 100
        X = np.tile(np.sin(2*np.pi * dom_freq *
                           np.arange(0, 1, 1.0 / sr)), (3, 1)).T
        result = accel.spectrum_features(
            X, sr=100, freq_range=None, prev_spectrum_features=None)
        np.testing.assert_array_almost_equal(
            result[0],
            [[1] * 3 +
             [4.283017e-01] * 3 +
             [6.107265e-01] * 3 +
             [0] * 3 +
             [0] * 3 +
             [4.283017e-01 / 6.107265e-01] * 3 +
             [1] * 3 +
             [4.283017e-01] * 3 +
             [np.nan] * 3 +
             [0.15508315] * 3]
        )
        np.testing.assert_array_equal(result[1], names)


class TestStats:
    def test_mean(self):
        # test on a single row
        X = np.array([[1., 1., 1., ]])
        result = accel.mean(X)
        np.testing.assert_array_equal(result[0], X)
        np.testing.assert_array_equal(
            result[1], ['MEAN_0', 'MEAN_1', 'MEAN_2'])

        # test on a single row with nan
        X = np.array([[1., np.nan, 1., ]])
        result = accel.mean(X)
        np.testing.assert_array_equal(result[0], X)
        np.testing.assert_array_equal(
            result[1], ['MEAN_0', 'MEAN_1', 'MEAN_2'])

        # test on an array
        X = np.array([[1., 1., 1., ], [2., 2., 2.]])
        result = accel.mean(X)
        np.testing.assert_array_equal(result[0], np.array([[1.5, 1.5, 1.5]]))
        np.testing.assert_array_equal(
            result[1], ['MEAN_0', 'MEAN_1', 'MEAN_2'])

        # test on an array with nan
        X = np.array([[1., 1., 1., ], [2., np.nan, 2.]])
        result = accel.mean(X)
        np.testing.assert_array_equal(result[0], np.array([[1.5, 1, 1.5]]))
        np.testing.assert_array_equal(
            result[1], ['MEAN_0', 'MEAN_1', 'MEAN_2'])

    def test_median(self):
        # test on a single row
        X = np.array([[1., 1., 1., ]])
        result = accel.median(X)
        np.testing.assert_array_equal(result[0], X)
        np.testing.assert_array_equal(
            result[1], ['MEDIAN_0', 'MEDIAN_1', 'MEDIAN_2'])

        # test on a single row with nan
        X = np.array([[1., np.nan, 1., ]])
        result = accel.median(X)
        np.testing.assert_array_equal(result[0], X)
        np.testing.assert_array_equal(
            result[1], ['MEDIAN_0', 'MEDIAN_1', 'MEDIAN_2'])

        # test on an array
        X = np.array([[1., 1., 1., ], [2., 2., 2.]])
        result = accel.median(X)
        np.testing.assert_array_equal(result[0], np.array([[1.5, 1.5, 1.5]]))
        np.testing.assert_array_equal(
            result[1], ['MEDIAN_0', 'MEDIAN_1', 'MEDIAN_2'])

        # test on larger array
        X = np.array([[1., 1., 1., ], [2., 2., 2.], [5., 5., 5.]])
        result = accel.median(X)
        np.testing.assert_array_equal(result[0], np.array([[2, 2, 2]]))
        np.testing.assert_array_equal(
            result[1], ['MEDIAN_0', 'MEDIAN_1', 'MEDIAN_2'])

        # test on an array with nan
        X = np.array([[1., 1., 1., ], [2., np.nan, 2.]])
        result = accel.median(X)
        np.testing.assert_array_equal(result[0], np.array([[1.5, 1, 1.5]]))
        np.testing.assert_array_equal(
            result[1], ['MEDIAN_0', 'MEDIAN_1', 'MEDIAN_2'])

    def test_std(self):
        # test on a single row
        X = np.array([[1., 1., 1., ]])
        result = accel.std(X)
        np.testing.assert_array_equal(result[0], np.array(
            [[np.nan, np.nan, np.nan, ]]))
        np.testing.assert_array_equal(result[1], ['STD_0', 'STD_1', 'STD_2'])

        # test on a single row with nan
        X = np.array([[1., np.nan, 1., ]])
        result = accel.std(X)
        np.testing.assert_array_equal(result[0], np.array(
            [[np.nan, np.nan, np.nan]]))
        np.testing.assert_array_equal(result[1], ['STD_0', 'STD_1', 'STD_2'])

        # test on an array
        X = np.array([[1., 1., 1., ], [2., 2., 2.]])
        result = accel.std(X)
        np.testing.assert_array_equal(result[0], np.array(
            [[np.sqrt(0.5), np.sqrt(0.5), np.sqrt(0.5)]]))
        np.testing.assert_array_equal(result[1], ['STD_0', 'STD_1', 'STD_2'])

        # test on an array with nan
        X = np.array([[1., 1., 1., ], [2., np.nan, 2.]])
        result = accel.std(X)
        np.testing.assert_array_equal(result[0], np.array(
            [[np.sqrt(0.5), np.nan, np.sqrt(0.5)]]))
        np.testing.assert_array_equal(result[1], ['STD_0', 'STD_1', 'STD_2'])

        # test on an array with more than two rows with nan
        X = np.array([[1., 1., 1., ], [2., 2, 2.], [1.5, np.nan, 1.5]])
        result = accel.std(X)
        np.testing.assert_array_equal(result[0], np.array(
            [[np.sqrt(0.25), np.sqrt(0.5), np.sqrt(0.25)]]))
        np.testing.assert_array_equal(result[1], ['STD_0', 'STD_1', 'STD_2'])

    def test_skew(self):
        # test on a single row
        X = np.array([[1., 1., 1., ]])
        result = accel.skew(X)
        np.testing.assert_array_equal(result[0], np.atleast_2d([0, 0, 0]))
        np.testing.assert_array_equal(
            result[1], ['SKEW_0', 'SKEW_1', 'SKEW_2'])

        # test on a single row with nan
        X = np.array([[1., np.nan, 1., ]])
        result = accel.skew(X)
        np.testing.assert_array_equal(result[0], np.atleast_2d([0, np.nan, 0]))
        np.testing.assert_array_equal(
            result[1], ['SKEW_0', 'SKEW_1', 'SKEW_2'])

        # test on an array
        X = np.array([[1., 1., 1., ], [2., 2., 2.]])
        result = accel.skew(X)
        np.testing.assert_array_equal(result[0], np.array([[0, 0, 0]]))
        np.testing.assert_array_equal(
            result[1], ['SKEW_0', 'SKEW_1', 'SKEW_2'])

        # test on normal distribution
        X = np.random.randn(1000000, 3)
        result = accel.skew(X)
        np.testing.assert_array_almost_equal(
            result[0], np.array([[0, 0, 0]]), decimal=1)
        np.testing.assert_array_equal(
            result[1], ['SKEW_0', 'SKEW_1', 'SKEW_2'])

        # test on normal distribution with nan
        X = np.random.randn(1000000, 3)
        X[np.random.randint(0, 49, 1), 0:3] = np.nan
        result = accel.skew(X)
        np.testing.assert_array_almost_equal(
            result[0], np.array([[0, 0, 0]]), decimal=1)
        np.testing.assert_array_equal(
            result[1], ['SKEW_0', 'SKEW_1', 'SKEW_2'])

        # test on exponential distribution
        X = np.random.standard_exponential(size=(1000000, 3))
        result = accel.skew(X)
        np.testing.assert_array_almost_equal(
            result[0], np.array([[2, 2, 2]]), decimal=1)
        np.testing.assert_array_equal(
            result[1], ['SKEW_0', 'SKEW_1', 'SKEW_2'])

    def test_kurtosis(self):
        # test on a single row
        X = np.array([[1., 1., 1., ]])
        result = accel.kurtosis(X)
        np.testing.assert_array_equal(result[0], np.atleast_2d([-3, -3, -3]))
        np.testing.assert_array_equal(
            result[1], ['KURTOSIS_0', 'KURTOSIS_1', 'KURTOSIS_2'])

        # test on a single row with nan
        X = np.array([[1., np.nan, 1., ]])
        result = accel.kurtosis(X)
        np.testing.assert_array_equal(
            result[0], np.atleast_2d([-3, np.nan, -3]))
        np.testing.assert_array_equal(
            result[1], ['KURTOSIS_0', 'KURTOSIS_1', 'KURTOSIS_2'])

        # test on uniform distribution
        X = np.random.uniform(low=0, high=1, size=(1000000, 3))
        result = accel.kurtosis(X)
        np.testing.assert_array_almost_equal(
            result[0], np.array([[-1.2, -1.2, -1.2]]), decimal=2)
        np.testing.assert_array_equal(
            result[1], ['KURTOSIS_0', 'KURTOSIS_1', 'KURTOSIS_2'])

        # test on normal distribution
        X = np.random.randn(1000000, 3)
        result = accel.kurtosis(X)
        np.testing.assert_array_almost_equal(
            result[0], np.array([[0, 0, 0]]), decimal=1)
        np.testing.assert_array_equal(
            result[1], ['KURTOSIS_0', 'KURTOSIS_1', 'KURTOSIS_2'])

        # test on normal distribution with nan
        X = np.random.randn(1000000, 3)
        X[np.random.randint(0, 49, 1), 0:3] = np.nan
        result = accel.kurtosis(X)
        np.testing.assert_array_almost_equal(
            result[0], np.array([[0, 0, 0]]), decimal=1)
        np.testing.assert_array_equal(
            result[1], ['KURTOSIS_0', 'KURTOSIS_1', 'KURTOSIS_2'])

        # test on laplace distribution
        X = np.random.laplace(loc=0, scale=1, size=(1000000, 3))
        result = accel.kurtosis(X)
        np.testing.assert_array_almost_equal(
            result[0], np.array([[3, 3, 3]]), decimal=1)
        np.testing.assert_array_equal(
            result[1], ['KURTOSIS_0', 'KURTOSIS_1', 'KURTOSIS_2'])

    def test_max_value(self):
        # test on a single row
        X = np.array([[1., 1., 1., ]])
        result = accel.max_value(X)
        np.testing.assert_array_equal(result[0], X)
        np.testing.assert_array_equal(result[1], ['MAX_0', 'MAX_1', 'MAX_2'])

        # test on a single row with nan
        X = np.array([[1., np.nan, 1., ]])
        result = accel.max_value(X)
        np.testing.assert_array_equal(result[0], X)
        np.testing.assert_array_equal(result[1], ['MAX_0', 'MAX_1', 'MAX_2'])

        # test on an array
        X = np.array([[1., 1., 1., ], [2., 2., 2.]])
        result = accel.max_value(X)
        np.testing.assert_array_equal(result[0], np.array([[2, 2, 2]]))
        np.testing.assert_array_equal(result[1], ['MAX_0', 'MAX_1', 'MAX_2'])

        # test on an array with nan
        X = np.array([[1., 1., 1., ], [2., np.nan, 2.]])
        result = accel.max_value(X)
        np.testing.assert_array_equal(result[0], np.array([[2, 1, 2]]))
        np.testing.assert_array_equal(result[1], ['MAX_0', 'MAX_1', 'MAX_2'])

    def test_min_value(self):
        # test on a single row
        X = np.array([[-1., -1., -1., ]])
        result = accel.min_value(X)
        np.testing.assert_array_equal(result[0], X)
        np.testing.assert_array_equal(result[1], ['MIN_0', 'MIN_1', 'MIN_2'])

        # test on a single row with nan
        X = np.array([[1., np.nan, 1., ]])
        result = accel.min_value(X)
        np.testing.assert_array_equal(result[0], X)
        np.testing.assert_array_equal(result[1], ['MIN_0', 'MIN_1', 'MIN_2'])

        # test on an array
        X = np.array([[1., 1., 1., ], [2., 2., 2.]])
        result = accel.min_value(X)
        np.testing.assert_array_equal(result[0], np.array([[1, 1, 1]]))
        np.testing.assert_array_equal(result[1], ['MIN_0', 'MIN_1', 'MIN_2'])

        # test on an array with nan
        X = np.array([[1., 1., 1., ], [2., np.nan, 2.]])
        result = accel.min_value(X)
        np.testing.assert_array_equal(result[0], np.array([[1, 1, 1]]))
        np.testing.assert_array_equal(result[1], ['MIN_0', 'MIN_1', 'MIN_2'])

    def test_range(self):
        # test on a single row
        X = np.array([[-1., -1., -1.]])
        result = accel.max_minus_min(X)
        np.testing.assert_array_equal(result[0], np.array([[0., 0., 0.]]))
        np.testing.assert_array_equal(
            result[1], ['RANGE_0', 'RANGE_1', 'RANGE_2'])

        # test on a single row with nan
        X = np.array([[1., np.nan, 1., ]])
        result = accel.max_minus_min(X)
        np.testing.assert_array_equal(result[0], np.array(
            [[0., np.nan, 0., ]]))
        np.testing.assert_array_equal(
            result[1], ['RANGE_0', 'RANGE_1', 'RANGE_2'])

        # test on an array
        X = np.array([[1., 1., 1., ], [3., 3., 3.]])
        result = accel.max_minus_min(X)
        np.testing.assert_array_equal(result[0], np.array([[2, 2, 2]]))
        np.testing.assert_array_equal(
            result[1], ['RANGE_0', 'RANGE_1', 'RANGE_2'])

        # test on an array with nan
        X = np.array([[1., 1., 1., ], [3., np.nan, 3.]])
        result = accel.max_minus_min(X)
        np.testing.assert_array_equal(result[0], np.array([[2, 0, 2]]))
        np.testing.assert_array_equal(
            result[1], ['RANGE_0', 'RANGE_1', 'RANGE_2'])

    def test_abs_max_value(self):
        # test on a single row
        X = np.array([[1., -1., 1., ]])
        result = accel.abs_max_value(X)
        np.testing.assert_array_equal(result[0], np.array([[1., 1., 1., ]]))
        np.testing.assert_array_equal(
            result[1], ['ABS_MAX_0', 'ABS_MAX_1', 'ABS_MAX_2'])

        # test on a single row with nan
        X = np.array([[1., np.nan, -1., ]])
        result = accel.abs_max_value(X)
        np.testing.assert_array_equal(result[0], np.array(
            [[1., np.nan, 1., ]]))
        np.testing.assert_array_equal(
            result[1], ['ABS_MAX_0', 'ABS_MAX_1', 'ABS_MAX_2'])

        # test on an array
        X = np.array([[1., 1., -1., ], [-2., 2., 2.]])
        result = accel.abs_max_value(X)
        np.testing.assert_array_equal(result[0], np.array([[2, 2, 2]]))
        np.testing.assert_array_equal(
            result[1], ['ABS_MAX_0', 'ABS_MAX_1', 'ABS_MAX_2'])

        # test on an array with nan
        X = np.array([[1., -1., 1., ], [-2., np.nan, 2.]])
        result = accel.abs_max_value(X)
        np.testing.assert_array_equal(result[0], np.array([[2, 1, 2]]))
        np.testing.assert_array_equal(
            result[1], ['ABS_MAX_0', 'ABS_MAX_1', 'ABS_MAX_2'])

    def test_abs_min_value(self):
        # test on a single row
        X = np.array([[-1., -1., -1.]])
        result = accel.abs_min_value(X)
        np.testing.assert_array_equal(result[0], np.array([[1., 1., 1.]]))
        np.testing.assert_array_equal(
            result[1], ['ABS_MIN_0', 'ABS_MIN_1', 'ABS_MIN_2'])

        # test on a single row with nan
        X = np.array([[-1., np.nan, -1., ]])
        result = accel.abs_min_value(X)
        np.testing.assert_array_equal(result[0], np.array(
            [[1., np.nan, 1., ]]))
        np.testing.assert_array_equal(
            result[1], ['ABS_MIN_0', 'ABS_MIN_1', 'ABS_MIN_2'])

        # test on an array
        X = np.array([[1., -1., 1., ], [-2., 2., 2.]])
        result = accel.abs_min_value(X)
        np.testing.assert_array_equal(result[0], np.array([[1, 1, 1]]))
        np.testing.assert_array_equal(
            result[1], ['ABS_MIN_0', 'ABS_MIN_1', 'ABS_MIN_2'])

        # test on an array with nan
        X = np.array([[1., -1., 1., ], [2., np.nan, -2.]])
        result = accel.abs_min_value(X)
        np.testing.assert_array_equal(result[0], np.array([[1, 1, 1]]))
        np.testing.assert_array_equal(
            result[1], ['ABS_MIN_0', 'ABS_MIN_1', 'ABS_MIN_2'])

    def test_zcr(self):
        # test on a single row
        X = np.array([[-1., -1., -1.]])
        result = accel.zcr(X)
        np.testing.assert_array_equal(
            result[0], np.array([[np.nan, np.nan, np.nan]]))
        np.testing.assert_array_equal(
            result[1], ['ZCR_0', 'ZCR_1', 'ZCR_2'])

        # test on a single row with nan
        X = np.array([[-1., np.nan, -1., ]])
        result = accel.zcr(X)
        np.testing.assert_array_equal(result[0], np.array(
            [[np.nan, np.nan, np.nan, ]]))
        np.testing.assert_array_equal(
            result[1], ['ZCR_0', 'ZCR_1', 'ZCR_2'])

        # test on an array
        X = np.array([[1., -1., 1., ], [-2., 2., 2.]])
        result = accel.zcr(X)
        np.testing.assert_array_equal(result[0], np.array([[1, 1, 0]]))
        np.testing.assert_array_equal(
            result[1], ['ZCR_0', 'ZCR_1', 'ZCR_2'])

        # test on an array with nan
        X = np.array([[1., -1., 1., ], [2., np.nan, -2.]])
        result = accel.zcr(X)
        np.testing.assert_array_equal(result[0], np.array([[0, 0, 1]]))
        np.testing.assert_array_equal(
            result[1], ['ZCR_0', 'ZCR_1', 'ZCR_2'])

    def test_mcr(self):
        # test on a single row
        X = np.array([[-1., -1., -1.]])
        result = accel.mcr(X)
        np.testing.assert_array_equal(
            result[0], np.array([[np.nan, np.nan, np.nan]]))
        np.testing.assert_array_equal(
            result[1], ['MCR_0', 'MCR_1', 'MCR_2'])

        # test on a single row with nan
        X = np.array([[-1., np.nan, -1., ]])
        result = accel.mcr(X)
        np.testing.assert_array_equal(result[0], np.array(
            [[np.nan, np.nan, np.nan, ]]))
        np.testing.assert_array_equal(
            result[1], ['MCR_0', 'MCR_1', 'MCR_2'])

        # test on an array
        X = np.array([[1., -1., 1., ], [-2., 2., 2.]])
        result = accel.mcr(X)
        np.testing.assert_array_equal(result[0], np.array([[1, 1, 1]]))
        np.testing.assert_array_equal(
            result[1], ['MCR_0', 'MCR_1', 'MCR_2'])

        # test on an array with nan
        X = np.array([[1., -1., 1., ], [2., np.nan, -2.]])
        result = accel.mcr(X)
        np.testing.assert_array_equal(result[0], np.array([[1, 0, 1]]))
        np.testing.assert_array_equal(
            result[1], ['MCR_0', 'MCR_1', 'MCR_2'])

    def test_correlation(self):
        # test on a single row
        X = np.array([[1., 1., 1.]])
        result = accel.correlation(X)
        np.testing.assert_array_equal(
            result[0], np.array([[np.nan, np.nan, np.nan]]))
        np.testing.assert_array_equal(
            result[1], ['CORRELATION_0', 'CORRELATION_1', 'CORRELATION_2'])

        # test on a single row with nan
        X = np.array([[-1., np.nan, -1., ]])
        result = accel.correlation(X)
        np.testing.assert_array_equal(result[0], np.array(
            [[np.nan, np.nan, np.nan]]))
        np.testing.assert_array_equal(
            result[1], ['CORRELATION_0', 'CORRELATION_1', 'CORRELATION_2'])

        # test on an array
        x = np.transpose(np.atleast_2d(np.arange(10)))
        y = np.transpose(-np.atleast_2d(np.arange(10)))
        z = np.transpose(np.atleast_2d(np.arange(10)/5))
        X = np.hstack((x, y, z))
        result = accel.correlation(X)
        np.testing.assert_array_almost_equal(
            result[0], np.array([[-1, 1, -1]]))
        np.testing.assert_array_equal(
            result[1], ['CORRELATION_0', 'CORRELATION_1', 'CORRELATION_2'])

        # test on an array with nan
        x = np.transpose(np.atleast_2d(np.arange(10)))
        y = np.transpose(-np.atleast_2d(np.arange(10)))
        z = np.transpose(np.atleast_2d(np.arange(10)/5))
        X = np.hstack((x, y, z))
        X[np.random.randint(0, 10, 1), 0:3] = np.nan
        result = accel.correlation(X)
        np.testing.assert_array_equal(
            result[0], np.array([[np.nan, np.nan, np.nan]]))
        np.testing.assert_array_equal(
            result[1], ['CORRELATION_0', 'CORRELATION_1', 'CORRELATION_2'])



"""
Module of extension functions to be applied to pandas objects (e.g., DataFrame or Series)

Author: Qu Tang

Date: 2020-02-03

License: see LICENSE file
"""
import functools
import pandas as pd
import numpy as np
import tqdm


def merge_all(*dfs, suffix_names, suffix_cols, **kwargs):
    def _append_suffix(df, suffix_name):
        new_cols = []
        for col in df.columns:
            if col in suffix_cols and suffix_name != '':
                col = col + '_' + suffix_name
            new_cols.append(col)
        df.columns = new_cols
        return df

    def _combine(left, right):
        left_df = _append_suffix(left[0], left[1])
        right_df = _append_suffix(right[0], right[1])
        merged = left_df.merge(
            right_df, **kwargs)
        return (merged, '')

    sequence = zip(dfs, suffix_names)
    if len(suffix_names) == 1:
        merged = _append_suffix(dfs[0], suffix_names[0])
    else:
        tuple_results = functools.reduce(_combine, sequence)
        merged = tuple_results[0]
    cols_with_suffixes = list(filter(lambda name: name.split('_')
                                     [-1] in suffix_names, merged.columns))
    return merged, cols_with_suffixes


def filter_column(df, col, values_to_filter_out=[]):
    # remove values
    is_valid_values = ~df[col].isin(values_to_filter_out).values
    filtered_df = df.loc[is_valid_values, :]
    return filtered_df


def parallel_apply(df, func, **kwargs):
    from pathos import pools
    import os
    import numpy as np
    cores = os.cpu_count()
    data_split = np.array_split(df, cores)
    pool = pools.ProcessPool(cores - 4)
    apply_func = functools.partial(func, **kwargs)
    data = pd.concat(pool.map(apply_func, data_split))
    pool.close()
    pool.join()
    return data


def fast_series_map(s, func, **kwargs):
    def _map(value):
        result[result == value] = func(value, **kwargs)
    result = s.copy()
    values = s.unique().tolist()
    [_map(value) for value in values]
    return result


def segment_by_time(df, seg_st=None, seg_et=None, st_col=0, et_col=None):
    et_col = et_col or st_col
    seg_st = seg_st or df.iloc[0, st_col]
    seg_et = seg_et or df.iloc[-1, et_col]
    if st_col == et_col:
        mask = (df.iloc[:, st_col] >= seg_st) & (
            df.iloc[:, et_col] < seg_et)
        return df.loc[mask, :].copy(deep=True)
    else:
        mask = (df.iloc[:, st_col] <= seg_et) & (
            df.iloc[:, et_col] >= seg_st)
        subset_df = df[mask].copy(deep=True)

        st_col = df.columns[st_col]
        et_col = df.columns[et_col]

        subset_df.loc[subset_df.loc[:, st_col] <
                      seg_st, st_col] = seg_st
        subset_df.loc[subset_df.loc[:, et_col] >
                      seg_et, et_col] = seg_et
        return subset_df


def get_common_timespan(*dfs, st=None, et=None, st_col=0, et_col=None):
    et_col = et_col or st_col

    if st is None:
        sts = [df.iloc[0, st_col] for df in dfs]
        st = pd.Timestamp(np.min(sts))
    else:
        st = pd.Timestamp(st)
    if et is None:
        ets = [df.iloc[-1, et_col] for df in dfs]
        et = pd.Timestamp(np.max(ets))
    else:
        et = pd.Timestamp(et)
    return st, et


def split_into_windows(*dfs, step_size, st=None, et=None, st_col=0, et_col=None):
    st, et = get_common_timespan(
        *dfs, st=st, et=et, st_col=st_col, et_col=et_col)
    step_size = step_size * 1000
    window_start_markers = pd.date_range(
        start=st, end=et, freq=f'{step_size}ms', closed='left')
    return window_start_markers


def fixed_window_slider(*dfs, slider_fn, window_size, step_size=None, st=None, et=None, st_col=0, et_col=None, show_progress=True, **slider_fn_kwargs):
    step_size = step_size or window_size
    window_start_markers = split_into_windows(
        *dfs, step_size=step_size, st=st, et=et, st_col=st_col, et_col=et_col)
    feature_sets = []
    if show_progress:
        bar = tqdm.tqdm(total=len(window_start_markers))

    result_dfs = []
    for window_st in window_start_markers:
        window_et = window_st + pd.Timedelta(window_size, unit='s')
        chunks = [segment_by_time(
            df, seg_st=window_st, seg_et=window_et, st_col=st_col, et_col=et_col) for df in dfs]
        result_df = slider_fn(
            *chunks, st=window_st, et=window_et, **slider_fn_kwargs)
        result_dfs.append(result_df)
        if show_progress:
            bar.set_description(
                f'Computed features for window: {window_st}')
            bar.update()
    if show_progress:
        bar.close()
    result_df = pd.concat(
        result_dfs, axis=0, ignore_index=True, sort=False)
    return result_df



import os


def is_large_file(filepath, threshold=20):
    size_in_bytes = os.path.getsize(filepath)
    size_in_mb = size_in_bytes / 1024 / 1024
    if filepath.endswith('gz'):
        threshold = int(threshold / 7)
    if size_in_mb > threshold:
        return True
    else:
        return False



from .. import mhealth_format as mh

MAPPING = {
    'dominant wrist': mh.DOMINANT_WRIST,
    'non dominant wrist': mh.NON_DOMINANT_WRIST,
    'dominant ankle': mh.DOMINANT_ANKLE,
    'non dominant ankle': mh.NON_DOMINANT_ANKLE,
    'dominant hip': mh.DOMINANT_HIP,
    'non dominant hip': mh.NON_DOMINANT_HIP,
    'dominant thigh': mh.DOMINANT_THIGH
}


def get_sensor_placement(dataset_path, pid, sid):
    p = mh.get_sensor_placement(dataset_path, pid, sid)
    return MAPPING[p]



"""
Pipeline with single sensor generator stream
=====================================================================

This example demonstrates using pipeline with a single sensor generator stream.
"""

# %%
# Imports
# ----------

import pandas as pd
import multiprocessing

from arus.core.pipeline import Pipeline
import arus
import datetime as dt

multiprocessing.freeze_support()

# %%
# pipeline processor test function
# ---------------------------------


def _pipeline_test_processor(chunk_list, **kwargs):
    import pandas as pd
    result = {'NAME': [],
              'START_TIME': [], 'STOP_TIME': []}
    for data, st, et, prev_st, prev_et, name in chunk_list:
        result['NAME'].append(name)
        result['START_TIME'].append(data.iloc[0, 0])
        result['STOP_TIME'].append(data.iloc[-1, 0])
    result = pd.DataFrame.from_dict(result)
    return result


# %%
# Setup stream
# --------------
window_size = 12.8
gr1 = arus.generator.RandomAccelDataGenerator(
    sr=80, grange=8, sigma=1, buffer_size=100)
seg1 = arus.segmentor.SlidingWindowSegmentor(window_size)
stream1 = arus.Stream(gr1, seg1, name='sensor-stream')

# %%
# Setup pipeline
# ----------------
# Here we use `threads` scheduler for demostration due to limitation of sphinx_gallery. In practice, you should better use `processes` scheduler to get benefit from multi-core processing.
pipeline = Pipeline(max_processes=1, scheduler='threads')
pipeline.add_stream(stream1)
pipeline.set_processor(_pipeline_test_processor)

# %%
# Start pipeline and read in processed data
# -------------------------------------------
pipeline.start()
results = []
for result, st, et, prev_st, prev_et, name in pipeline.get_iterator():
    result['PREV_WINDOW_ST'] = prev_st
    results.append(result)
    if len(results) == 10:
        break

# %%
# Stop pipeline
# ----------------
success = pipeline.stop()

# %%
# Output of the processed data
# ------------------------------
output = pd.concat(results, axis=0, sort=False)
output



import enum
from loguru import logger
import queue
import time

import PySimpleGUI as sg

import app_state as app
import backend
import base
import components as comp

from arus.models.muss import Strategy


class Event(enum.Enum):
    START_TASK = enum.auto()
    PROGRESS_UPDATE = enum.auto()
    TASK_COMPLETED = enum.auto()
    RESULT_READY = enum.auto()
    PLACEMENT_SELECTED = enum.auto()


class Mode(enum.Enum):
    TRAIN = enum.auto()
    LOSO = enum.auto()


class TaskWindow(base.BaseWindow):
    def __init__(self, title, strategies=[Strategy.USE_ORIGIN_ONLY], mode=Mode.TRAIN):
        super().__init__(title, 100)
        self._strategies = strategies
        self._progress_queue = queue.Queue()
        self._task = None
        self._result = None
        self._selected_strategy = None
        self._selected_placements = None
        self._mode = mode

    def init_states(self):
        state = app.AppState.getInstance()
        self._selected_strategy = self._selected_strategy or self._strategies[0]
        self._progress_text = "Task is not started..."
        self._timer = comp.Timer()
        if self._mode == Mode.TRAIN:
            if len(self._strategies) == 1:
                self._result = state.origin_model
            else:
                self._result = state.new_model
        elif self._mode == Mode.LOSO:
            if len(self._strategies) == 1:
                self._result = state.origin_validate_results
            else:
                self._result = state.new_validate_results
        if self._result is not None:
            self._events.put(Event.RESULT_READY)
        return state

    def init_views(self):
        header_row = [
            comp.heading(
                self._title,
                fixed_column_width=self._column_width
            )
        ]

        self._start_button = comp.control_button(
            "Start task",
            disabled=False,
            key=Event.START_TASK
        )
        start_button_row = [
            self._start_button
        ]

        placements = ['DW', 'DA', 'DT']
        placement_list_row = [
            comp.selection_list(placements, default_selections=placements,
                                mode='multiple', fixed_column_width=self._column_width, rows=len(placements), key=Event.PLACEMENT_SELECTED)
        ]

        strategy_row = self.init_strategy_controls()

        self._progress_bar = comp.ProgressBar(
            self._progress_text,
            fixed_column_width=self._column_width
        )
        progress_text_row = self._progress_bar.get_component()[0]
        progress_bar_row = self._progress_bar.get_component()[1]

        if self._mode == Mode.TRAIN:
            result_view = self.init_train_view()
        elif self._mode == Mode.LOSO:
            result_view = self.init_loso_view()
        else:
            result_view = []
        layout = [
            header_row,
            start_button_row,
            placement_list_row,
            strategy_row,
            progress_text_row,
            progress_bar_row,
            result_view
        ]

        return sg.Window(
            self._title,
            layout=layout,
            finalize=True
        )

    def _update_states_and_events(self, event, values):
        self._selected_placements = values[Event.PLACEMENT_SELECTED]
        for strategy in self._strategies:
            if values[strategy]:
                self._selected_strategy = strategy
                break
        try:
            progress = self._progress_queue.get(timeout=0.02)
            if type(progress) is str:
                self._events.put(Event.PROGRESS_UPDATE)
                self._progress_text = progress
            else:
                self._task = progress
                self._events.put(Event.PROGRESS_UPDATE)
            self._timer.tick()
        except queue.Empty:
            pass
        if self._task is not None:
            since_last_tick = self._timer.since_last_tick()
            if not self._task.ready() and since_last_tick >= 1:
                self._events.put(Event.PROGRESS_UPDATE)
                self._timer.tick()
            elif self._task.ready():
                self._timer.tick()
                self._result = self._task.get()
                self._events.put(Event.TASK_COMPLETED)
                self._progress_text = 'Completed in ' + \
                    self._timer.get_total_lapsed_time(
                        formatted=True) + ' seconds.'
                self._task = None
                logger.info("Training lapsed time: " +
                            self._timer.get_total_lapsed_time(formatted=True))

    def _dispatch_events(self, event):
        if event == Event.START_TASK:
            self.disable_start_button()
            self._timer.start()
            self.start_task()
        elif event == Event.PROGRESS_UPDATE:
            self.increment_progress()
        elif event == Event.TASK_COMPLETED:
            self._timer.stop()
            self._timer.reset()
            self.increment_progress(1.0)
            self.enable_start_button()
            self.close_task_pool()
            if self._mode == Mode.TRAIN:
                self.display_model_summary()
            elif self._mode == Mode.LOSO:
                self.display_confusion_matrix()
                self.display_classification_report()
        elif event == Event.RESULT_READY:
            if self._mode == Mode.TRAIN:
                self.display_model_summary()
            elif self._mode == Mode.LOSO:
                self.display_confusion_matrix()
                self.display_classification_report()

    def init_strategy_controls(self):
        default_strategy_index = self._strategies.index(
            self._selected_strategy)
        self._control_strategies = comp.RadioGroup(
            'STRATEGY',
            len(self._strategies),
            default_index=default_strategy_index,
            fixed_column_width=self._column_width
        )
        strategy_names = [strategy.name for strategy in self._strategies]
        self._control_strategies.add_radioboxes(
            strategy_names,
            keys=self._strategies,
            disable_states=[False] * len(self._strategies)
        )
        return self._control_strategies.get_component()

    def init_train_view(self):
        summary = backend.get_model_summary(self._result)
        self._text_model_summary = comp.text(
            summary, fixed_column_width=self._column_width)
        summary_row = [
            self._text_model_summary
        ]
        return summary_row

    def init_loso_view(self):
        self._view_cm = comp.Plot(
            fixed_column_width=int(self._column_width / 2), rows=10)
        self._view_report = comp.table(cells=[["", "", "", ""]], headings=[
                                       "class", "precision", "recall", "f1-score"], fixed_column_width=int(self._column_width / 2))
        return [self._view_cm.get_component(), self._view_report]

    def disable_start_button(self):
        self._start_button.update('Training...', disabled=True)

    def enable_start_button(self):
        self._start_button.update('Start training', disabled=False)

    def start_task(self):
        if self._mode == Mode.TRAIN:
            backend.train_model(
                origin_labels=self._state.origin_labels,
                origin_dataset=self._state.origin_dataset,
                new_dataset=self._state.new_dataset,
                new_labels=self._state.new_labels,
                progress_queue=self._progress_queue,
                strategy=self._selected_strategy,
                pool=self._state.task_pool,
                placement_names=self._selected_placements
            )
        elif self._mode == Mode.LOSO:
            backend.validate_model(
                origin_labels=self._state.origin_labels,
                origin_dataset=self._state.origin_dataset,
                new_dataset=self._state.new_dataset,
                new_labels=self._state.new_labels,
                progress_queue=self._progress_queue,
                strategy=self._selected_strategy,
                pool=self._state.task_pool,
                placement_names=self._selected_placements
            )

    def increment_progress(self, percentage=None):
        if percentage is None:
            self._progress_bar.increment(text=self._progress_text)
        else:
            self._progress_bar.update(percentage, self._progress_text)

    def display_model_summary(self):
        summary = backend.get_model_summary(self._result)
        self._text_model_summary.update(summary)

    def display_confusion_matrix(self):
        result = self._result
        fig = self._view_cm.get_figure()
        if fig is None:
            fig = backend.get_confusion_matrix_figure(result)
            self._view_cm.update_plot(fig)
        else:
            backend.get_confusion_matrix_figure(result, fig)
            self._view_cm.update_plot()

    def display_classification_report(self):
        result = self._result
        values = backend.get_classification_report_table(result)
        self._view_report.Update(values=values, visible=True,
                                 num_rows=min(len(values), 10))

    def close_task_pool(self):
        self._state.task_pool.close()
        self._state.task_pool.join()

    def get_result(self):
        return self._result



import queue
import threading
import time
import enum
from loguru import logger
import typing


class Stream:
    """
    The base class for data stream.

    Stream class is an abstraction of any data source that can be loaded into memory in arbitrary chunk size either asynchronously (currently only support threading) or synchronously.
    """

    class Status(enum.Enum):
        """Stream Status codes."""
        NOT_START = enum.auto()
        START = enum.auto()
        RUN = enum.auto()
        STOP = enum.auto()

    def __init__(self, generator: "arus.generator.Generator", segmentor: "arus.segmentor.Segmentor", name: typing.Optional[str] = 'default-stream', scheduler: typing.Optional[str] = 'thread'):
        """

        Arguments:
            generator: a Generator instance that can provide streaming data.
            segmentor: a Segmentor instance that is responsible for segmenting the streaming data into chunks.
            name: The name of the data stream will also be used as the name of the sub-thread that is used to load data. Defaults to 'default-stream'.
            scheduler: The scheduler used to load the data source. It can be either 'thread' or 'sync'. Defaults to 'thread'.
        """
        self._input_buffer = queue.Queue()
        self._output_buffer = queue.Queue()
        self._status = Stream.Status.NOT_START
        self._name = name
        self._scheduler = scheduler
        self._generator = generator
        self._segmentor = segmentor

    def generate(self) -> "pandas.Dataframe":
        """A python generator function to get the segmented streaming data.

        Returns:
            segmented streaming data.
        """
        while True:
            data = self._output_buffer.get()
            if data is None:
                self._status = Stream.Status.STOP
                break
            yield data

    def start(self, start_time: "str, datetime, numpy.datetime64, pandas.Timestamp" = None):
        """Method to start loading data from the provided data source.

        Arguments:
            start_time: The reference time for segmentation.

        Note:
            `start_time` is used to sync between multiple streams. If it is `None`, the default value would be extracted from the first sample of the loaded data.
        """
        self._status = Stream.Status.START
        logger.info('Stream is starting.')
        self._segmentor.set_ref_time(start_time)
        self._loading_thread = self._get_thread_for_loading()
        self._loading_thread.daemon = True
        self._segment_thread = self._get_thread_for_chunking()
        self._segment_thread.daemon = True
        self._loading_thread.start()
        self._segment_thread.start()
        while not self._loading_thread.isAlive() or not self._segment_thread.isAlive():
            time.sleep(0.1)
        logger.info('Stream started.')
        self._status = Stream.Status.RUN

    def stop(self):
        """Stop the loading process."""
        logger.info('Stream is stopping.')
        self._status = Stream.Status.STOP
        self._segmentor.stop()
        self._generator.stop()
        logger.info('Segmentor thread stopped.')
        time.sleep(0.1)
        self._loading_thread.join(timeout=1)
        logger.info('Generator thread stopped.')
        with self._output_buffer.mutex:
            self._output_buffer.queue.clear()
        with self._input_buffer.mutex:
            self._input_buffer.queue.clear()
        self._status = Stream.Status.NOT_START
        logger.info('Stream stopped.')

    def get_status(self) -> "Stream.Status":
        """Get the status code of the stream.

        Returns:
            The status code of the stream.
        """
        return self._status

    def _get_thread_for_loading(self):
        return threading.Thread(
            target=self._generate, name=self._name + '-loading')

    def _get_thread_for_chunking(self):
        return threading.Thread(
            target=self._segment, name=self._name + '-segmenting')

    def _generate(self):
        logger.info('Generator thread started.')
        self._generator.run()
        logger.info('Generator thread is stopping.')

    def _segment(self):
        logger.info('Segmentor thread started.')
        while True:
            try:
                data, _ = next(self._generator.get_result())
                if data is None:
                    break
                for segment in self._segmentor.segment(data):
                    if self._status == Stream.Status.STOP:
                        break
                    self._output_buffer.put(segment + (self._name, ))
            except:
                pass
            finally:
                if self._status == Stream.Status.STOP:
                    break
        self._output_buffer.put(None)
        logger.info('Segmentor thread is stopping.')



from sklearn.model_selection import TimeSeriesSplit


class TimeSeriesEpisodeSplit:
    def __init__(self, fcs_df, fs_names, task_name, n_split=5):
        self.fs_names = fs_names
        self.task_name = task_name
        self.episodes = fcs_df.groupby(task_name)
        self.spliters = {}
        self.n_split = n_split
        for name, episode_df in self.episodes:
            X = episode_df.index
            spliter = TimeSeriesSplit(n_splits=self.n_split)
            self.spliters[name] = spliter.split(X)

    def split(self, X=None, y=None, groups=None):
        for i in range(self.n_split):
            train_idx = []
            test_idx = []
            for name, episode_df in self.episodes:
                train, test = next(self.spliters[name])
                train = episode_df.index[train]
                test = episode_df.index[test]
                train_idx += list(train)
                test_idx += list(test)
            train_idx = sorted(train_idx)
            test_idx = sorted(test_idx)
            yield train_idx, test_idx



"""
Module of util functions to process date and time

Author: Qu Tang

Date: 2020-02-03

License: see LICENSE file
"""

import numpy as np
import pandas as pd
import datetime as dt
import time
import tzlocal
import pytz


class Moment:
    def __init__(self, obj):
        """Date time object that unifies various date time types

        Args:
            obj (int, float, datetime, np.datetime64, pd.Timestamp): If the obj is a datetime, np.datetime64 or pd.Timestamp object and does not have time zone, it will be assumed to be local time.

        Raises:
            NotImplementedError: When input argument type is not recognized.
        """
        if type(obj) == int or type(obj) == float:
            self._posix = float(obj)
        elif type(obj) == str:
            # if it is string, always ignore time zone
            ts = pd.Timestamp(obj).tz_localize(None)
            self._posix = obj.to_pydatetime().timestamp()
        elif type(obj) == dt.datetime:
            # assume local time, always to posix time
            self._posix = obj.timestamp()
        elif type(obj) == np.datetime64:
            # assume local time, so convert to utc first
            local_ts = obj.astype(float) / 10 ** 6
            self._posix = local_ts + Moment.get_local_to_utc_offset(local_ts)
        elif type(obj) == pd.Timestamp:
            self._posix = obj.to_pydatetime().timestamp()
        else:
            raise NotImplementedError('Input type is not supported')

    def to_unix_timestamp(self):
        return self._posix

    def to_pandas_timestamp(self, tz=None):
        if tz is not None:
            return pd.Timestamp(self._posix, unit='s', tz=tz)
        else:
            return pd.Timestamp.fromtimestamp(self._posix)

    def to_datetime(self, tz=None):
        if tz is not None:
            return dt.datetime.fromtimestamp(self._posix, tz=tz)
        else:
            return dt.datetime.fromtimestamp(self._posix)

    def to_datetime64(self):
        return np.datetime64(self.to_datetime())

    def to_string(self, fmt, tz=None):
        obj = self.to_datetime(tz=tz)
        return obj.strftime(fmt)

    @staticmethod
    def get_local_to_utc_offset(ts=None):
        ts = ts or time.time()
        return (dt.datetime.utcfromtimestamp(ts) - dt.datetime.fromtimestamp(ts)
                ).total_seconds()

    @staticmethod
    def get_utc_to_local_offset(ts=None):
        return - Moment.get_local_to_utc_offset(ts=ts)

    @staticmethod
    def get_local_timezone():
        return tzlocal.get_localzone()

    @staticmethod
    def get_utc_timezone():
        return pytz.timezone('UTC')

    @staticmethod
    def get_timezone(obj):
        return pytz.timezone(obj)

    @staticmethod
    def get_duration(st, et, unit='s'):
        st = Moment(st).to_datetime()
        et = Moment(et).to_datetime()
        if unit == 's':
            unit_len = dt.timedelta(seconds=1)
        elif unit == 'ms':
            unit_len = dt.timedelta(milliseconds=1)
        elif unit == 'us':
            unit_len = dt.timedelta(microseconds=1)
        elif unit == 'm':
            unit_len = dt.timedelta(minutes=1)
        return (et - st) / unit_len

    @staticmethod
    def get_durations(st, et, unit='s'):
        return [Moment.get_duration(s, e) for s, e in zip(st, et)]

    @staticmethod
    def transform(iterable):
        return [Moment(item) for item in iterable]

    @staticmethod
    def get_sequence(st, sr, N, endpoint_as_extra=True, tz=None, format='pandas'):
        if endpoint_as_extra:
            N = N + 1
        st = Moment(st).to_unix_timestamp()
        interval = 1 / float(sr)
        span = interval * (N - 1)
        et = st + span
        ts = np.linspace(start=st, stop=et, num=N, endpoint=True).tolist()
        if format == 'pandas':
            ts = [Moment(t).to_pandas_timestamp(tz=tz) for t in ts]
        elif format == 'posix':
            pass
        elif format == 'datetime':
            ts = [Moment(t).to_datetime(tz=tz) for t in ts]
        return ts

    @staticmethod
    def seq_to_unix_timestamp(seq, fmt=None):
        if type(seq[0]) in [int, float, np.float64, np.int64]:
            return seq
        elif type(seq[0]) == str:
            seq = pd.to_datetime(seq, format=fmt)
        elif type(seq[0]) == dt.datetime:
            local_ts = pd.to_datetime(seq).values.astype(float) / 10 ** 9
            local_ts = local_ts + Moment.get_local_to_utc_offset(local_ts[0])
            return local_ts
        elif type(seq[0]) == np.datetime64:
            local_ts = np.array(seq).astype(
                'datetime64[us]').astype(float) / 10 ** 6
            local_ts = local_ts + Moment.get_local_to_utc_offset(local_ts[0])
            return local_ts
        elif type(seq[0]) == pd.Timestamp:
            local_ts = pd.to_datetime(seq).values.astype(float) / 10 ** 9
            local_ts = local_ts + Moment.get_local_to_utc_offset(local_ts[0])
            return local_ts
        else:
            raise NotImplementedError('Input type is not supported')



"""
Functions to manipulate dataset stored in mhealth format.

Author: Qu Tang

Date: 01/31/2020

License: GNU v3
"""

from .constants import *
from .helper import *
from .core import *
from .io import *



from ..plugins import actigraph
from .. import testing
from .. import mhealth_format as mh
from .. import dataset
import numpy as np
import pandas as pd
import pytest


@pytest.fixture(params=['actigraph_imu', 'actigraph_accel'])
def test_file(request):
    return dataset.get_sample_datapath(request.param)


class TestActigraphReader:
    @pytest.mark.parametrize('chunksize', [None, 5])
    def test_read(self, test_file, chunksize):
        reader = actigraph.ActigraphReader(
            test_file, has_ts=True, has_header=True)
        reader.read(chunksize=chunksize)
        results = []
        for data in reader.get_data():
            results.append(data)
        if chunksize is None:
            assert len(results) == 1
        else:
            assert len(results) > 1
        if 'imu' in test_file:
            assert results[0].shape[1] == 11
        else:
            assert results[0].shape[1] == 4
        assert results[0].columns[0] == mh.constants.TIMESTAMP_COL

        meta = reader.get_meta()
        assert type(meta) == dict
        assert len(meta['VERSION_CODE']) == 6
        assert meta['SAMPLING_RATE'] > 30
        assert meta['SENSOR_ID'].startswith('TAS')
        if 'imu' in test_file:
            assert meta['IMU']
            assert meta['DYNAMIC_RANGE'] == 16
        else:
            assert not meta['IMU']
            assert meta['DYNAMIC_RANGE'] == 8


class TestActigraphSensorFileGenerator:
    @pytest.mark.parametrize('buffer_size', [None, 5])
    def test_generate(self, test_file, buffer_size):
        gr = actigraph.ActigraphSensorFileGenerator(
            test_file, buffer_size=buffer_size)
        sizes = []
        gr.run()
        for data, _ in gr.get_result():
            if data is None:
                break
            assert type(data) == pd.DataFrame
            sizes.append(data.shape[0])
        if buffer_size is None:
            assert len(sizes) == 1
            assert sizes[0] > 5
        else:
            sizes = sizes[:-1]
            assert np.all(np.array(sizes) == 5)



import pytest
from .. import synchronizer
import time
import numpy as np


class TestSynchronizer:
    def test_single_source(self):
        time0 = time.time()
        times = [time0 + i for i in range(0, 5)]
        sync = synchronizer.Synchronizer()
        sync.add_source()

        for st in times:
            sync.run(1, src=None, context={
                'start_time': st, 'stop_time': st + 2, 'data_id': 'test'})
        sts = []
        for data, context in sync.get_result():
            assert data is not None
            np.testing.assert_array_equal(data[0], [1])
            assert 'data_ids' in context
            sts.append(context['start_time'])
        np.testing.assert_array_equal(sts, times)

    def test_multiple_sources(self):
        time0 = time.time()
        times = [time0 + i for i in range(0, 10)]
        sync = synchronizer.Synchronizer()
        sync.add_sources(2)
        # in order one by one alternatively
        for st in times:
            sync.run(1, src=None, context={
                'start_time': st, 'stop_time': st + 2, 'data_id': 'test1'})
            sync.run(2, src=None, context={
                'start_time': st, 'stop_time': st + 2, 'data_id': 'test2'})
        sts = []
        for data, context in sync.get_result():
            np.testing.assert_array_equal(data, [1, 2])
            assert 'data_ids' in context
            sts.append(context['start_time'])
        np.testing.assert_array_equal(sts, times)

        # source 1 adds all at first
        for st in times:
            sync.run(1, src=None, context={
                'start_time': st, 'stop_time': st + 2, 'data_id': 'test1'})
        for st in times:
            sync.run(2, src=None, context={
                'start_time': st, 'stop_time': st + 2, 'data_id': 'test2'})
        sts = []
        for data, context in sync.get_result():
            np.testing.assert_array_equal(data, [1, 2])
            assert 'data_ids' in context
            sts.append(context['start_time'])
        np.testing.assert_array_equal(sts, times)

        # source 2 has different time order
        for st in times:
            sync.run(1, src=None, context={
                'start_time': st, 'stop_time': st + 2, 'data_id': 'test1'})
        for st in times[5:] + times[:5]:
            sync.run(2, src=None, context={
                'start_time': st, 'stop_time': st + 2, 'data_id': 'test2'})
        sts = []
        for data, context in sync.get_result():
            np.testing.assert_array_equal(data, [1, 2])
            assert 'data_ids' in context
            sts.append(context['start_time'])
        np.testing.assert_array_equal(sts, times[5:] + times[:5])

    def test_dynamic_added_sources(self):
        time0 = time.time()
        times = [time0 + i for i in range(0, 10)]
        sync = synchronizer.Synchronizer()
        sync.add_source()
        # in order one by one alternatively
        for st in times[:5]:
            sync.run(1, src=None, context={
                'start_time': st, 'stop_time': st + 2, 'data_id': 'test1'})
        sts = []
        for data, context in sync.get_result():
            np.testing.assert_array_equal(data, [1])
            assert 'data_ids' in context
            sts.append(context['start_time'])
        np.testing.assert_array_equal(sts, times[:5])

        sync.add_source()
        # source 1 adds all at first
        for st in times[5:8]:
            sync.run(1, src=None, context={
                'start_time': st, 'stop_time': st + 2, 'data_id': 'test1'})
        for st in times[5:8]:
            sync.run(2, src=None, context={
                'start_time': st, 'stop_time': st + 2, 'data_id': 'test2'})
        sts = []
        for data, context in sync.get_result():
            np.testing.assert_array_equal(data, [1, 2])
            assert 'data_ids' in context
            sts.append(context['start_time'])
        np.testing.assert_array_equal(sts, times[5:8])
        # source 2 has different time order
        for st in times[8:]:
            sync.run(1, src=None, context={
                'start_time': st, 'stop_time': st + 2, 'data_id': 'test1'})
            sync.run(2, src=None, context={
                'start_time': st, 'stop_time': st + 2, 'data_id': 'test2'})
        sts = []
        for data, context in sync.get_result():
            np.testing.assert_array_equal(data, [1, 2])
            assert 'data_ids' in context
            sts.append(context['start_time'])
        np.testing.assert_array_equal(sts, times[8:])



"""
Computing features related to activation samples.

Author: Qu Tang

Date: Jul 10, 2018

References:

1. Mannini A, Rosenberger M, Haskell WL, Sabatini AM, Intille SS. Activity
 Recognition in Youth Using Single Accelerometer Placed at Wrist or Ankle. Med
 Sci Sports Exerc. 2017;49(4):801–12.
 https://www.ncbi.nlm.nih.gov/pubmed/27820724

"""

import numpy as np
from .. import extensions as ext


ACTIVATION_FEATURE_NAME_PREFIX = [
    'ACTIVE_SAMPLES', 'ACTIVATIONS', 'MEAN_ACTIVATION_DURATIONS', 'STD_ACTIVATION_DURATIONS'
]


def _active_samples(X, threshold=0.2):
    active_samples = X >= threshold
    return active_samples


def _stats_active_samples_1d(active_samples):
    active_edges = np.diff(active_samples, prepend=0, append=0)
    rising_edges = np.argwhere(active_edges > 0)
    falling_edges = np.argwhere(active_edges < 0)
    durations = []
    if len(rising_edges) == 0:
        durations = [0, 0]
    else:
        if rising_edges[0] > falling_edges[0]:
            falling_edges = falling_edges[1:]
        if rising_edges[-1] > falling_edges[-1]:
            rising_edges = rising_edges[0:-1]
        if len(rising_edges) == 0:
            durations = [0, 0]
    num_activation_crossings = len(rising_edges)
    for r, f in zip(rising_edges, falling_edges):
        durations.append(f - r)
    mean_activation_duration = np.mean(durations)
    if len(durations) <= 1:
        std_activation_duration = 0
    else:
        std_activation_duration = np.std(durations, ddof=1)
    return np.array([num_activation_crossings, mean_activation_duration, std_activation_duration])


def _stats_active_samples(active_samples):
    result = np.apply_along_axis(
        _stats_active_samples_1d, axis=0, arr=active_samples)
    return result


def activation_features(X,
                        threshold=0.2,   selected=ACTIVATION_FEATURE_NAME_PREFIX):
    X = ext.numpy.atleast_float_2d(X)
    W = X.shape[0]

    fv = []
    fv_names = []

    active_samples = _active_samples(X, threshold=threshold)
    num_active_samples = np.sum(active_samples, axis=0, keepdims=True)
    stats = _stats_active_samples(active_samples)

    """
    This feature identifies the amount of activity within the window that is over the threshold, thereby providing a rough estimate of the amount of relevant activity being recorded in the window. This feature could distinguish between activities that result in relevant acceleration in most of the window and activities in which the relevant activity takes place for only a portion of the window.
    """
    if ACTIVATION_FEATURE_NAME_PREFIX[0] in selected:
        fv.append(num_active_samples / W)
        fv_names += [
            f'{ACTIVATION_FEATURE_NAME_PREFIX[0]}_{i}' for i in range(X.shape[1])]
    """
    "This feature identifies the number of threshold crossings within the window (rising edges only), normalized to the number of active samples FS, thereby capturing movement fragmentation within the window. This feature could discriminate impulsive events from longer-lasting acceleration events, since it quantifies how many times within the window the acceleration passed from the inactive to the active condition."
    """
    if ACTIVATION_FEATURE_NAME_PREFIX[1] in selected:
        num_activations = np.atleast_2d(stats[0, :])
        fv.append(np.divide(
            num_activations, num_active_samples, where=num_activations != 0.0))
        fv_names += [
            f'{ACTIVATION_FEATURE_NAME_PREFIX[1]}_{i}' for i in range(X.shape[1])]
    """
    This captures the mean duration of activation intervals within the window, normalized by the window length. An activation interval is defined as the amount of samples between two consecutive threshold crossings. This feature provides information on movement bout fragmentation within the window that could help discriminate between activities that involve stable movements, such as those in natural walking, and those with short bouts, such as sport ones.
    """
    if ACTIVATION_FEATURE_NAME_PREFIX[2] in selected:
        mean_activation_durations = np.atleast_2d(stats[1, :])
        fv.append(mean_activation_durations / W)
        fv_names += [
            f'{ACTIVATION_FEATURE_NAME_PREFIX[2]}_{i}' for i in range(X.shape[1])]
    """
    This feature captures the standard deviation of the duration of activation intervals within the window, normalized by the window length, thereby providing information on uniformity of activation intervals within the window. This feature may help discriminate between activities with cyclic movements with a very stable ratio between active and inactive phases, and more random activities. A stable cyclic movement would result in lower variability of activation intervals that would repeat themselves within the window. Fast and aperiodic movements, however, such as those in recreational activities would result in highly variable activation bouts.
    """
    if ACTIVATION_FEATURE_NAME_PREFIX[3] in selected:
        std_activation_durations = np.atleast_2d(stats[2, :])
        fv.append(std_activation_durations / W)
        fv_names += [
            f'{ACTIVATION_FEATURE_NAME_PREFIX[3]}_{i}' for i in range(X.shape[1])]

    if len(fv) == 0:
        return None, None

    result = np.concatenate(fv, axis=1)

    return result, fv_names



import PySimpleGUI as sg
import app_state as app
import dashboard
import backend
import os
from loguru import logger
import traceback
import arus


def control_button(text, disabled, key=None):
    return sg.Button(button_text=text,
                     font=('Helvetica', 15), auto_size_button=True, size=(25, None), key=key, disabled=disabled)


class SessionSelectionPanel:
    def __init__(self):
        self._new_button = None
        self._continue_button = None

    def init_panel(self):
        self._new_button = control_button(
            'New session', disabled=False, key='_NEW_')
        self._continue_button = control_button(
            'Continue last session', disabled=not os.path.exists(app.AppState._snapshot_path), key='_CONTINUE_'
        )
        return sg.Window('Select a session to start', layout=[
            [self._new_button, self._continue_button]
        ], finalize=True)

    def start(self):
        window = self.init_panel()
        ready = False
        while True:
            event, _ = window.read()
            if event == self._continue_button.Key:
                logger.info('Restoring application status..')
                file_path = sg.PopupGetFile('Select the pkl file to restore session', title='Continue a session',
                                            default_extension='.pkl', initial_folder=app.AppState._snapshot_path)
                if file_path is None:
                    continue
                if app.AppState.restore(file_path):
                    app_state = app.AppState.getInstance()
                    ready = True
                else:
                    app_state = app.AppState.getInstance()
                    app_state.origin_dataset = backend.load_origin_dataset()
                    ready = True
            elif event == self._new_button.Key:
                app.AppState.reset()
                app_state = app.AppState.getInstance()
                new_pid = sg.PopupGetText(
                    'Set new participant ID',
                    title='Create a new session',
                    default_text=app_state.pid,
                    keep_on_top=True
                )
                if new_pid is None:
                    continue
                app_state.pid = new_pid
                app_state.origin_dataset = backend.load_origin_dataset()
                ready = True
            elif event is None:
                break
            if ready:
                try:
                    log_file = os.path.join(
                        app.AppState._path, 'logs', app_state.pid + '.log')
                    os.makedirs(os.path.dirname(log_file), exist_ok=True)
                    logger.add(log_file)
                    demo = dashboard.Dashboard(
                        title='Arus Demo Session: ' + app_state.pid)
                    demo.start()
                except Exception as e:
                    print(e)
                    print(traceback.format_exc())
                finally:
                    logger.info('Saving application status..')
                    app.AppState.snapshot()
        window.close()



# %%

import arus
from loguru import logger
arus.dev.set_default_logger()

# %% [markdown]
# ## input operators

# %% [markdown]
# ### generator operator

# %%
generator = arus.generator.RandomAccelDataGenerator(sr=80, buffer_size=10)
op = arus.Node(generator, name='accel', t=arus.Node.Type.INPUT)

op.start()
i = 0
while True:
    data = next(op.produce())
    if data.signal == arus.Node.Signal.WAIT:
        continue
    logger.info(data)
    if i == 2:
        break
    i += 1
op.stop()

# %% [markdown]
# ### metawear generator operator

# %%
scanner = arus.plugins.metawear.MetaWearScanner()
addrs = scanner.get_nearby_devices(max_devices=1)
generator = arus.plugins.metawear.MetaWearAccelDataGenerator(
    addrs[0], sr=50, grange=8, buffer_size=10)
op = arus.Node(generator, name=addrs[0], t=arus.Node.Type.INPUT)

op.start()
i = 0
while True:
    data = next(op.produce())
    if data.signal == arus.Node.Signal.WAIT:
        continue
    logger.info(data)
    if i == 2:
        break
    i += 1
op.stop()

# %% [markdown]
# ### mhealth file generator operator

# %%
spades_lab = arus.dataset.load_dataset('spades_lab')
mhealth_filepath = spades_lab['subjects']['SPADES_2']['sensors']['DW'][0]
generator = arus.generator.MhealthSensorFileGenerator(
    mhealth_filepath, buffer_size=10)

op = arus.Node(generator, name='SPADES_2-DW', t=arus.Node.Type.INPUT)

op.start()
i = 0
while True:
    data = next(op.produce())
    if data.signal == arus.Node.Signal.WAIT:
        continue
    logger.info(data)
    if i == 2:
        break
    i += 1
op.stop()

# %% [markdown]
# ## pipe operators

# %% [markdown]
# ### segmentor operator

# %%
spades_lab = arus.dataset.load_dataset('spades_lab')
mhealth_filepath = spades_lab['subjects']['SPADES_2']['sensors']['DW'][0]
generator = arus.generator.MhealthSensorFileGenerator(
    mhealth_filepath, buffer_size=1800)
generator.run()
values, context = next(generator.get_result())
generator.stop()
segmentor = arus.segmentor.SlidingWindowSegmentor(window_size=1)
op = arus.Node(segmentor, name='segmentor-1s')

op.start()
op.consume(arus.Node.Pack(values=values, signal=arus.Node.Signal.DATA,
                          context=context, src='SPADES_2-dw'))
i = 0
while True:
    data = next(op.produce())
    if data.signal == arus.Node.Signal.WAIT:
        continue
    logger.info(data)
    if i == 2:
        break
    i += 1
op.stop()


# %% [markdown]
# ### synchronizer operator


# %%
spades_lab = arus.dataset.load_dataset('spades_lab')
dw_filepath = spades_lab['subjects']['SPADES_2']['sensors']['DW'][0]
da_filepath = spades_lab['subjects']['SPADES_2']['sensors']['DA'][0]
dw_generator = arus.generator.MhealthSensorFileGenerator(
    dw_filepath, buffer_size=1800)
da_generator = arus.generator.MhealthSensorFileGenerator(
    da_filepath, buffer_size=1800)
dw_generator.run()
da_generator.run()
dw_values, dw_context = next(dw_generator.get_result())
da_values, da_context = next(da_generator.get_result())
dw_context = {**dw_context, 'data_id': 'DW'}
da_context = {**da_context, 'data_id': 'DA'}
dw_generator.stop()
da_generator.stop()

segmentor = arus.segmentor.SlidingWindowSegmentor(window_size=1)
segmentor.run(
    dw_values, src=None, context=dw_context)
seg_dw_values, dw_context = next(segmentor.get_result())
segmentor = arus.segmentor.SlidingWindowSegmentor(window_size=1)
segmentor.run(
    da_values, src=None, context=da_context)
seg_da_values, da_context = next(segmentor.get_result())

synchronizer = arus.synchronizer.Synchronizer()
synchronizer.add_sources(2)
op = arus.Node(synchronizer, name='sync-dw-da')
op.start()
op.consume(arus.Node.Pack(values=seg_dw_values,
                          signal=arus.Node.Signal.DATA, context=dw_context, src='dw'))
op.consume(arus.Node.Pack(values=seg_da_values,
                          signal=arus.Node.Signal.DATA, context=da_context, src='da'))
while True:
    data = next(op.produce())
    if data.signal == arus.Node.Signal.WAIT:
        continue
    logger.info(data)
    break
op.stop()


# %% [markdown]
# ### scheduler/processor operator


# %%
spades_lab = arus.dataset.load_dataset('spades_lab')
dw_filepath = spades_lab['subjects']['SPADES_2']['sensors']['DW'][0]
da_filepath = spades_lab['subjects']['SPADES_2']['sensors']['DA'][0]
dw_generator = arus.generator.MhealthSensorFileGenerator(
    dw_filepath, buffer_size=1800)
dw_generator.run()
dw_values, dw_context = next(dw_generator.get_result())
dw_context = {**dw_context, 'data_id': 'DW'}
dw_generator.stop()


def compute_mean(values, src=None, **context):
    import numpy as np
    data = values.values[:, 1:]
    result = np.mean(data, axis=0)
    return result, context


processor = arus.processor.Processor(compute_mean,
                                     mode=arus.Scheduler.Mode.THREAD,
                                     scheme=arus.Scheduler.Scheme.SUBMIT_ORDER, max_workers=2)

op = arus.Node(processor, name='compute-mean')
op.start()
op.consume(arus.Node.Pack(values=dw_values, signal=arus.Node.Signal.DATA,
                          context=dw_context, src='dw'))
while True:
    data = next(op.produce())
    if data.signal == arus.Node.Signal.WAIT:
        continue
    logger.info(data)
    break
op.stop()



from ._model import *
from .splitter import *
from .report import *
import pkg_resources


def get_prebuilt(model_name):
    return pkg_resources.resource_filename('arus', f'models/prebuilt/{model_name}.har')



"""
Demonstration of the usage of arus.core.pipeline
=====================================================================

This example shows how to pass previous output and input to processor for a pipeline.
"""

from arus.core.pipeline import Pipeline
import arus
import datetime as dt
import pandas as pd


def _pipeline_test_processor(chunk_list, prev_input=None, prev_output=None, **kwargs):
    import pandas as pd
    result = {'NAME': [],
              'START_TIME': [], 'STOP_TIME': []}
    for data, st, et, prev_st, prev_et, name in chunk_list:
        result['NAME'].append(name)
        result['START_TIME'].append(data.iloc[0, 0])
        result['STOP_TIME'].append(data.iloc[-1, 0])
    result = pd.DataFrame.from_dict(result)
    if prev_output is not None:
        result['PREV_START_TIME'] = prev_output['START_TIME']
        result['PREV_STOP_TIME'] = prev_output['STOP_TIME']
    else:
        result['PREV_START_TIME'] = None
        result['PREV_STOP_TIME'] = None
    return result


if __name__ == "__main__":
    window_size = 12.8
    gr1 = arus.generator.RandomAccelDataGenerator(
        sr=80, grange=8, sigma=1, buffer_size=100)
    seg1 = arus.segmentor.SlidingWindowSegmentor(window_size)
    stream1 = arus.Stream(gr1, seg1, name='stream-1')

    pipeline = Pipeline(
        max_processes=2, scheduler='processes', preserve_status=True)
    pipeline.add_stream(stream1)
    pipeline.set_processor(_pipeline_test_processor)
    pipeline.start()
    results = []
    prev_results = []
    for result, st, et, prev_st, prev_et, name in pipeline.get_iterator():
        print(st)
        result['WINDOW_ST'] = st
        result['WINDOW_ET'] = et
        result['PREV_WINDOW_ST'] = prev_st
        result['PREV_WINDOW_ET'] = prev_et
        result['STREAM_NAME'] = name
        results.append(result)

        if len(results) == 10:
            break
    pipeline.stop()
    print(pd.concat(results, axis=0, sort=False))



"""ARUS dataflow: generators.

Generator functions can take external data source and generate values in buffer_size (number of samples) with Python generator pattern.

* Author: Qu Tang
* Date: 01/28/2020
* License: GNU v3
"""

import datetime as dt
import time

import numpy as np
import pandas as pd

from . import mhealth_format as mh
from . import moment, operator


class Generator(operator.Operator):
    """Base generator class.

    Note:
        This class should always be inherited and should not be called directly. Subclasses should override `run` method with its own data source.

    Args:
        buffer_size: The sample size for each burst of the streaming data.

    Example:
        Use generator classes in the following pattern.

        ```python
        # Replace Generator, args, kwargs with proper names for different generator classes
        gen = Generator(*args, **kwargs)
        # Start generator
        gen.start()
        # Get chunked data
        for data, context in gen.get_result():
            # handle data
            if data is None: # Some condition for early termination
                break
        # Stop generator
        gen.stop()
        ```
    """

    def __init__(self, buffer_size: int = 1000000):
        super().__init__()
        self._buffer_size = buffer_size
        self._buffer = None
        self._stop = False

    def start(self):
        """Generate burst of streaming data.

        Use this method instead of `run` when you are using generators directly instead of relying on `arus.Node`.
        """
        self.run()

    def run(self, values=None, src=None, context={}):
        """Implementation of data generation.

        **This method must be overrided by subclasses and developers should implement it to load data from data sources and generate chunks with the data.**


        Examples:
            Developers should implement with the following template.

            ```python
            # You can accept data source from the `__init__` method
            for data in self._load_data(self._data_sources):
                # Call this to buffer input data with `buffer_size`
                result = self._buffering(data)
                # Put the generated data into `self._result`. You should always attach the `self._context` so that it can be chained with other operators via `arus.Node`.
                self._result.put((result, self._context))
                # Implement stop condition
                if data is None or self._stop:
                    break
            ```

        Args:
            values: Not used.
            src: Not used.
            context: Not used.


        """
        pass

    def _buffering(self, data):
        if self._buffer_size is None:
            return data
        if self._buffer is None:
            if data.shape[0] == self._buffer_size:
                return data
            elif data.shape[0] < self._buffer_size:
                self._buffer = data
                return None
            else:
                self._buffer = data.iloc[self._buffer_size:, :]
                return data.iloc[:self._buffer_size, :]
        else:
            n = self._buffer_size - self._buffer.shape[0]
            self._buffer = pd.concat(
                (self._buffer, data.iloc[:n, :]), axis=0, sort=False)
            if self._buffer.shape[0] == self._buffer_size:
                result = self._buffer.copy()
                self._buffer = data.iloc[n:, :]
                return result
            else:
                return None


class MhealthSensorFileGenerator(Generator):
    """Generator class for sensor files stored in mhealth format.

    Note:
        The file paths should be sorted before loading. The generator will load data from the files one by one in order.

    Args:
        *filepaths: The sensor file paths as data sources.
        **kwargs: Other keyword arguments passed to parent class, which is `buffer_size`.

    Examples:
        Generate mhealth sensor data in chunks, with each chunk includes 10 samples.

        ```python
        gen = arus.MhealthSensorFileGenerator("path/to/sensor_file.csv", buffer_size=10)
        gen.start()
        for data, context in gen.get_result():
            print(data.shape[0]) # should be 10
            if data is None: # end condition
                break
        gen.stop()
        ```
    """

    def __init__(self, *filepaths: str, **kwargs: object):
        super().__init__(**kwargs)
        self._filepaths = filepaths

    def run(self, values=None, src=None, context={}):
        """Implementation of data generation (Hidden).
        """
        for filepath in self._filepaths:
            reader = mh.MhealthFileReader(filepath)
            reader.read_csv(chunksize=self._buffer_size)
            for data in reader.get_data():
                if self._stop:
                    break
                result = self._buffering(data)
                if result is not None:
                    self._result.put((result, self._context))
            if self._stop:
                break
        self._result.put((None, self._context))


class MhealthAnnotationFileGenerator(Generator):
    """Generator class for annotation files stored in mhealth format.

    Note:
        The file paths should be sorted before loading. The generator will load data from the files one by one in order.

    Args:
        *filepaths: The annotation file paths as data sources.
        **kwargs: Other keyword arguments passed to parent class, which is `buffer_size`.

    Examples:
        Generate mhealth annotation data in chunks, with each chunk includes 10 rows of annotations.

        ```python
        gen = arus.MhealthAnnotationFileGenerator("path/to/annotation_file.csv", buffer_size=10)
        gen.start()
        for data, context in gen.get_result():
            print(data.shape[0]) # should be 10
            if data is None: # end condition
                break
        gen.stop()
        ```
    """

    def __init__(self, *filepaths, **kwargs):
        super().__init__(**kwargs)
        self._filepaths = filepaths

    def run(self, values=None, src=None, context={}):
        """Implementation of data generation (Hidden).
        """
        for filepath in self._filepaths:
            reader = mh.MhealthFileReader(filepath)
            reader.read_csv(chunksize=self._buffer_size,
                            datetime_cols=[0, 1, 2])
            for data in reader.get_data():
                if self._stop:
                    break
                result = self._buffering(data)
                if result is not None:
                    self._result.put((result, self._context))
            if self._stop:
                break
        self._result.put((None, self._context))


class RandomAccelDataGenerator(Generator):
    """Generator class for raw accelerometer data synthesized randomly.

    Args:
        sr: The sampling rate in Hz.
        grange: The dynamic range in g value.
        st: The start timestamp of the generated data. If `None`, it will be the current timestamp.
        sigma: The variance of the generated data sampled from Gaussian Distribution.
        max_samples: The maximum number of samples to be generated. 
        **kwargs: Other keyword arguments passed to parent class, which is `buffer_size`.

    Examples:
        Generate accelerometer data in chunks, with each chunk includes 10 samples for at most 100 samples (10 chunks).

        ```python
        gen = arus.RandomAccelDataGenerator(80, grange=8, st=datetime.datetime.now(), sigma=1.5, max_samples=100, buffer_size=10)
        gen.start()
        for data, context in gen.get_result():
            print(data.shape[0]) # should be 10
            # should end loop after 10 cycles
        gen.stop()
        ```
    """

    def __init__(self, sr: int, grange: int = 8, st: "str, datetime, numpy.datetime64, pandas.Timestamp" = None, sigma: float = 1, max_samples: int = None, **kwargs: object):
        super().__init__(**kwargs)
        self._sr = sr
        self._grange = grange
        self._st = st or dt.datetime.now()
        self._sigma = sigma
        self._max_samples = max_samples
        self._max_count = max_samples or 1

    def run(self, values=None, src=None, context={}):
        """Implementation of data generation (Hidden).
        """
        counter = 0
        while counter <= self._max_count:
            if self._stop:
                break
            data = np.random.standard_normal(
                size=(self._buffer_size, 3)) * self._sigma
            data[data > self._grange] = self._grange
            data[data < -self._grange] = -self._grange
            ts = moment.Moment.get_sequence(
                self._st, self._sr, self._buffer_size, format='pandas')
            self._st = ts[-1]
            ts = ts[0:-1]
            result = pd.DataFrame(
                index=ts, data=data).reset_index(drop=False)
            result.columns = [mh.TIMESTAMP_COL, 'X', 'Y', 'Z']
            time.sleep(0.2)
            counter += self._buffer_size
            if self._max_samples is None:
                self._max_count = counter + 1
            self._result.put((result, self._context))
        self._result.put((None, self._context))


class RandomAnnotationDataGenerator(Generator):
    """Generator class for annotation data synthesized randomly.

    Args:
        labels: List of annotation labels to be randomly selected.
        duration_mu: The mean of the Gaussian distribution in seconds used to decide the annotation duration.
        duration_sigma: The standard deviation of the Gaussian distribution in seconds used to decide the annotation duration.
        num_mu: The mean of the Gaussian distribution used to decide the number of annotations for each generation.
        num_sigma: The standard deviation of the Gaussian distribution used to decide the number of annotations for each generation.
        st: The start timestamp of the generated data. If `None`, it will be the current timestamp.
        max_samples: The maximum number of rows of annotations to be generated.
        **kwargs: Other keyword arguments passed to parent class, which is `buffer_size`.

    Examples:
        Generate annotation data in chunks, with each chunk includes 10 samples for at most 100 samples (10 chunks).

        ```python
        gen = arus.RandomAnnotationDataGenerator(['Sit', 'Walk'], duration_mu=5, duration_sigma=5, st=st=datetime.datetime.now(), num_mu=3, num_sigma=1, max_samples=100, buffer_size=10)
        gen.start()
        for data, context in gen.get_result():
            print(data.shape[0]) # should be 10
            # should end loop after 10 cycles
        gen.stop()
        ```
    """

    def __init__(self, labels: list, duration_mu: float = 5, duration_sigma: float = 5, st: "str, datetime, numpy.datetime64, pandas.Timestamp" = None, num_mu: float = 2, num_sigma: float = 1, max_samples: int = None, **kwargs: object):
        super().__init__(**kwargs)
        self._labels = labels
        self._duration_mu = duration_mu
        self._duration_sigma = duration_sigma
        self._st = st or dt.datetime.now()
        self._num_mu = num_mu
        self._num_sigma = num_sigma
        self._max_samples = max_samples
        self._max_count = self._max_samples or 1

    def run(self, values=None, src=None, context={}):
        """Implementation of data generation (Hidden).
        """
        counter = 0
        while counter <= self._max_count:
            if self._stop:
                break
            N = np.random.poisson(lam=self._num_mu)
            durations = np.random.standard_normal(
                size=N) * self._duration_sigma + self._duration_mu
            start_times = [self._st]
            stop_times = []
            for duration in durations:
                new_start_time = self._st + pd.Timedelta(duration, 'S')
                start_times.append(new_start_time)
                self._st = new_start_time
                stop_times.append(new_start_time)
            start_times = start_times[:-1]
            label_names = np.random.choice(self._labels, N)
            result = pd.DataFrame.from_dict({
                mh.TIMESTAMP_COL: start_times,
                mh.START_TIME_COL: start_times,
                mh.STOP_TIME_COL: stop_times,
                mh.ANNOTATION_LABEL_COL: label_names
            })
            time.sleep(0.2)
            counter += N
            if self._max_samples is None:
                self._max_count = counter + 1
            self._result.put((result, self._context))
        self._result.put((None, self._context))



"""
Frequency domain features for numerical time series data
"""
import numpy as np
from loguru import logger
from scipy import signal

from .. import extensions as ext

SPECTRUM_FEATURE_NAME_PREFIX = [
    'DOM_FREQ',
    'DOM_FREQ_POWER',
    'TOTAL_FREQ_POWER',
    'FREQ_POWER_ABOVE_3DOT5',
    'FREQ_POWER_RATIO_ABOVE_3DOT5',
    'DOM_FREQ_POWER_RATIO',
    'FREQ_POWER_RATIO_BEWTEEN_DOT5_2DOT5',
    'DOM_FREQ_POWER_BETWEEN_DOT6_2DOT6',
    'DOM_FREQ_RATIO_PREV_BOUT',
    'SPECTRAL_ENTROPY'
]


def spectrum_features(X, sr, n=1, freq_range=None, prev_spectrum_features=None, selected=SPECTRUM_FEATURE_NAME_PREFIX):
    X = ext.numpy.atleast_float_2d(X)
    # fill nan at first, nan will be filled by spline interpolation
    X = ext.numpy.mutate_nan(X)
    freq, Sxx = _fft(X, sr, freq_range=freq_range)
    freq_peaks, Sxx_peaks = _fft_peaks(freq, Sxx)

    fv = []
    fv_names = []

    if SPECTRUM_FEATURE_NAME_PREFIX[0] in selected:
        fv.append(_dom_freq(
            freq_peaks, Sxx_peaks, n=n))
        fv_names += [
            f'{SPECTRUM_FEATURE_NAME_PREFIX[0]}_{i}' for i in range(X.shape[1])
        ]

    if SPECTRUM_FEATURE_NAME_PREFIX[1] in selected:
        fv.append(_dom_freq_power(
            freq_peaks, Sxx_peaks, n=n))
        fv_names += [
            f'{SPECTRUM_FEATURE_NAME_PREFIX[1]}_{i}' for i in range(X.shape[1])
        ]

    if SPECTRUM_FEATURE_NAME_PREFIX[2] in selected:
        fv.append(_total_freq_power(Sxx))
        fv_names += [
            f'{SPECTRUM_FEATURE_NAME_PREFIX[2]}_{i}' for i in range(X.shape[1])
        ]

    if SPECTRUM_FEATURE_NAME_PREFIX[3] in selected:
        fv.append(_freq_power_above_3_point_5(freq, Sxx))
        fv_names += [
            f'{SPECTRUM_FEATURE_NAME_PREFIX[3]}_{i}' for i in range(X.shape[1])
        ]

    if SPECTRUM_FEATURE_NAME_PREFIX[4] in selected:
        fv.append(_freq_power_ratio_above_3_point_5(freq, Sxx))
        fv_names += [
            f'{SPECTRUM_FEATURE_NAME_PREFIX[4]}_{i}' for i in range(X.shape[1])
        ]

    if SPECTRUM_FEATURE_NAME_PREFIX[5] in selected:
        fv.append(_dom_freq_power_ratio(
            freq, Sxx, freq_peaks, Sxx_peaks, n=n))
        fv_names += [
            f'{SPECTRUM_FEATURE_NAME_PREFIX[5]}_{i}' for i in range(X.shape[1])
        ]

    if SPECTRUM_FEATURE_NAME_PREFIX[6] in selected:
        fv.append(_freq_power_ratio_between_point_5_and_2_point_5(
            freq, Sxx
        ))
        fv_names += [
            f'{SPECTRUM_FEATURE_NAME_PREFIX[6]}_{i}' for i in range(X.shape[1])
        ]

    if SPECTRUM_FEATURE_NAME_PREFIX[7] in selected:
        fv.append(_dom_freq_power_between_point_6_and_2_point_6(
            freq_peaks,
            Sxx_peaks
        ))
        fv_names += [
            f'{SPECTRUM_FEATURE_NAME_PREFIX[7]}_{i}' for i in range(X.shape[1])]

    if SPECTRUM_FEATURE_NAME_PREFIX[8] in selected:
        if prev_spectrum_features is not None:
            fv.append(_dom_freq_ratio_previous_bout(
                freq_peaks, Sxx_peaks, prev_dom_freq=prev_spectrum_features[0, :], n=1))
        else:
            fv.append(_dom_freq_ratio_previous_bout(
                freq_peaks, Sxx_peaks, prev_dom_freq=None, n=1))
        fv_names += [
            f'{SPECTRUM_FEATURE_NAME_PREFIX[8]}_{i}' for i in range(X.shape[1])
        ]

    if SPECTRUM_FEATURE_NAME_PREFIX[9] in selected:
        fv.append(_spectral_entropy(freq, Sxx))
        fv_names += [
            f'{SPECTRUM_FEATURE_NAME_PREFIX[9]}_{i}' for i in range(X.shape[1])
        ]

    if len(fv) == 0:
        return None, None

    result = np.concatenate(fv, axis=1)
    return result, fv_names


def _fft(X, sr, freq_range=None):
    freq, time, Sxx = signal.spectrogram(
        X,
        fs=sr,
        window='hamming',
        nperseg=X.shape[0],
        noverlap=0,
        detrend='constant',
        return_onesided=True,
        scaling='density',
        axis=0,
        mode='magnitude')
    Sxx = np.abs(Sxx[:, :, 0])
    # interpolate to get values in the freq_range
    if freq_range is not None:
        Sxx_interpolated = np.interp(freq_range, freq, Sxx)
    else:
        Sxx_interpolated = Sxx
        freq_range = freq
    Sxx_interpolated = np.atleast_2d(Sxx_interpolated)
    return freq_range, Sxx_interpolated


def _sort_fft_peaks(freq, Sxx, i, j):
    if len(i) == 0:
        sorted_freq_peaks = np.array([0])
        sorted_Sxx_peaks = np.array([np.nanmean(Sxx, axis=0)[j]])
    else:
        freq_peaks = freq[i]
        Sxx_peaks = Sxx[i, j]
        sorted_i = np.argsort(Sxx_peaks)
        sorted_i = sorted_i[:: -1]
        sorted_freq_peaks = freq_peaks[sorted_i]
        sorted_Sxx_peaks = Sxx_peaks[sorted_i]
    # logger.debug('sxx:' + str(j) + ":" + str(sorted_Sxx_peaks.shape))
    # logger.debug('freq:' + str(j) + ":" + str(sorted_freq_peaks.shape))
    return (sorted_freq_peaks, sorted_Sxx_peaks)


def _fft_peaks(freq, Sxx):
    n_axis = Sxx.shape[1]
    m_freq = Sxx.shape[0]
    if m_freq == 1:
        freq_peaks = [freq] * n_axis
        Sxx_peaks = [Sxx[:, i] for i in range(Sxx.shape[1])]
    else:
        # at least 0.1 Hz different when looking for peak
        mpd = int(np.ceil(1.0 / (freq[1] - freq[0]) * 0.1))
        # print(self._Sxx.shape)
        # mph should not be set, because signal can be weak but there may still be some dominant frequency, 06/03/2019
        freq_peak_indices = list(map(lambda x: ext.numpy.detect_peaks(
            x, mph=None, mpd=mpd), list(Sxx.T)))
        # i = list(map(lambda x: detect_peaks(
        #     x, mph=1e-3, mpd=mpd), list(self._Sxx.T)))
        axes = range(0, n_axis)
        freq_peaks = []
        Sxx_peaks = []
        for i, j in zip(freq_peak_indices, axes):
            freq_peak, Sxx_peak = _sort_fft_peaks(freq, Sxx, i, j)
            freq_peaks.append(freq_peak)
            Sxx_peaks.append(Sxx_peak)
        # note that freq_peaks and Sxx_peaks will have structure
        # [freq_peaks_for_x_axis, freq_peaks_for_y_axis, ...]
        #  And each item would be a 1d array
    return freq_peaks, Sxx_peaks


def _spectral_entropy(freqs, Sxx):
    # normalized
    sum_of_Sxx = np.sum(Sxx, axis=0)
    psd = Sxx / sum_of_Sxx
    s_entropy = -np.sum(np.multiply(psd, np.log2(psd)), axis=0)
    result = s_entropy / np.log2(len(freqs))
    result = np.atleast_2d(result)
    return result


def _dom_freq(freq_peaks, Sxx_peaks, n=1):
    result = []
    for i in range(len(freq_peaks)):
        if len(freq_peaks[i]) >= n:
            if not np.isnan(Sxx_peaks[i][n-1]):
                result.append(freq_peaks[i][n-1])
            else:
                result.append(np.nan)
        else:
            result.append(np.nan)
    result = np.atleast_2d(np.array(result))
    return result


def _dom_freq_power(freq_peaks, Sxx_peaks, n=1):
    result = []
    for i in range(len(Sxx_peaks)):
        if len(Sxx_peaks[i]) >= n:
            result.append(Sxx_peaks[i][n-1])
        else:
            result.append(np.nan)
    result = np.atleast_2d(np.array(result))
    return result


def _total_freq_power(Sxx):
    result = np.sum(Sxx, axis=0, keepdims=True)
    return result


def _dom_freq_in_band(freq_peaks, Sxx_peaks, low=0, high=np.inf, n=1):
    result = []
    for i in range(len(Sxx_peaks)):
        freq = freq_peaks[i]
        indices = (freq >= low) & (freq <= high)
        limited_freq = freq[indices]
        if len(limited_freq) < n:
            result.append(np.nan)
        else:
            result.append(limited_freq[n-1])
    result = np.atleast_2d(np.array(result))
    return result


def _dom_freq_power_in_band(freq_peaks, Sxx_peaks, low=0, high=np.inf, n=1):
    result = []
    for i in range(len(Sxx_peaks)):
        freq = freq_peaks[i]
        Sxx = Sxx_peaks[i]
        indices = (freq >= low) & (freq <= high)
        limited_Sxx = Sxx[indices]
        if len(limited_Sxx) < n:
            result.append(np.nan)
        else:
            result.append(limited_Sxx[n-1])
    result = np.atleast_2d(np.array(result))
    return result


def _total_freq_power_in_band(freq, Sxx, low=0, high=np.inf):
    indices = (freq >= low) & (freq <= high)
    limited_Sxx = Sxx[indices, :]
    limited_total_power = np.sum(limited_Sxx, axis=0, keepdims=True)
    return limited_total_power


def _freq_power_above_3_point_5(freq, Sxx):
    result = _total_freq_power_in_band(freq, Sxx, low=3.5)
    return result


def _freq_power_ratio_above_3_point_5(freq, Sxx):
    highend_power = _freq_power_above_3_point_5(freq, Sxx)
    total_power = _total_freq_power(Sxx)
    result = np.divide(highend_power, total_power, out=np.zeros_like(
        total_power), where=total_power != 0)
    return result


def _freq_power_ratio_between_point_5_and_2_point_5(freq, Sxx):
    band_power = _total_freq_power_in_band(freq, Sxx, low=0.5, high=2.5)
    total_power = _total_freq_power(Sxx)
    result = np.divide(band_power, total_power, out=np.zeros_like(
        total_power), where=total_power != 0)
    return result


def _dom_freq_power_ratio(freq, Sxx, freq_peaks, Sxx_peaks, n=1):
    total_power = _total_freq_power(Sxx)
    dom_freq_power = _dom_freq_power(freq_peaks, Sxx_peaks, n=n)
    result = np.divide(dom_freq_power,
                       total_power, out=np.zeros_like(total_power), where=total_power != 0)
    return result


def _dom_freq_between_point_6_and_2_point_6(freq_peaks, Sxx_peaks):
    result = _dom_freq_in_band(freq_peaks, Sxx_peaks,
                               low=0.6, high=2.6, n=1)
    return result


def _dom_freq_power_between_point_6_and_2_point_6(freq_peaks, Sxx_peaks):
    result = _dom_freq_power_in_band(freq_peaks, Sxx_peaks, low=0.6, high=2.6,
                                     n=1)
    return result


def _dom_freq_ratio_previous_bout(freq_peaks, Sxx_peaks, prev_dom_freq=None, n=1):
    if prev_dom_freq is None:
        result = np.full((1, len(Sxx_peaks)), np.nan)
    else:
        current_dom_freq = _dom_freq(freq_peaks, Sxx_peaks, n=1)
        result = np.divide(current_dom_freq, np.atleast_2d(prev_dom_freq), out=np.zeros_like(
            prev_dom_freq), where=prev_dom_freq != 0)
    return result



import pytest
from .. import scheduler
import os
import time


def task1():
    print('task1 start on {}'.format(os.getpid()))
    time.sleep(2)
    print('task1 stop on {}'.format(os.getpid()))
    return 'task1', time.time(), os.getpid()


def task2():
    print('task2 start on {}'.format(os.getpid()))
    time.sleep(0.5)
    print('task2 stop on {}'.format(os.getpid()))
    return 'task2', time.time(), os.getpid()


def task3():
    print('task3 start on {}'.format(os.getpid()))
    time.sleep(1)
    print('task3 stop on {}'.format(os.getpid()))
    return 'task3', time.time(),  os.getpid()


class TestScheduler:
    @pytest.mark.parametrize('scheme', [scheduler.Scheduler.Scheme.SUBMIT_ORDER, scheduler.Scheduler.Scheme.EXECUTION_ORDER, scheduler.Scheduler.Scheme.AFTER_PREVIOUS_DONE])
    @pytest.mark.parametrize('mode', [scheduler.Scheduler.Mode.THREAD, scheduler.Scheduler.Mode.PROCESS])
    def test_modes_and_schemes_with_get_all_remaining_results(self, scheme, mode):
        sch = scheduler.Scheduler(mode=mode, scheme=scheme, max_workers=5)
        sch.submit(task1)
        sch.submit(task2)
        sch.submit(task3)
        results = sch.get_all_remaining_results()
        if mode == scheduler.Scheduler.Mode.THREAD:
            assert len(set([r[2] for r in results])) == 1
        else:
            assert len(set([r[2] for r in results])) == 3
        if scheme == scheduler.Scheduler.Scheme.EXECUTION_ORDER:
            assert results[0][0] == 'task2'
            assert results[1][0] == 'task3'
            assert results[2][0] == 'task1'
        else:
            assert results[0][0] == 'task1'
            assert results[1][0] == 'task2'
            assert results[2][0] == 'task3'
        if scheme == scheduler.Scheduler.Scheme.AFTER_PREVIOUS_DONE:
            assert results[2][1] > results[1][1]
            assert results[1][1] > results[0][1]
        elif scheme == scheduler.Scheduler.Scheme.EXECUTION_ORDER:
            assert results[2][1] > results[1][1]
            assert results[1][1] > results[0][1]
        else:
            assert results[0][1] > results[2][1]
            assert results[2][1] > results[1][1]
        sch.shutdown()

    @pytest.mark.parametrize('scheme', [scheduler.Scheduler.Scheme.SUBMIT_ORDER, scheduler.Scheduler.Scheme.EXECUTION_ORDER, scheduler.Scheduler.Scheme.AFTER_PREVIOUS_DONE])
    @pytest.mark.parametrize('mode', [scheduler.Scheduler.Mode.THREAD, scheduler.Scheduler.Mode.PROCESS])
    def test_modes_and_schemes_with_get_result(self, scheme, mode):
        sch = scheduler.Scheduler(mode=mode, scheme=scheme, max_workers=5)
        sch.submit(task1)
        sch.submit(task2)
        sch.submit(task3)
        results = []
        while True:
            try:
                result = sch.get_result()
                results.append(result)
            except scheduler.Scheduler.ResultNotAvailableError:
                continue
            if len(results) == 3:
                break
        if mode == scheduler.Scheduler.Mode.THREAD:
            assert len(set([r[2] for r in results])) == 1
        else:
            assert len(set([r[2] for r in results])) == 3
        if scheme == scheduler.Scheduler.Scheme.EXECUTION_ORDER:
            assert results[0][0] == 'task2'
            assert results[1][0] == 'task3'
            assert results[2][0] == 'task1'
        else:
            assert results[0][0] == 'task1'
            assert results[1][0] == 'task2'
            assert results[2][0] == 'task3'
        if scheme == scheduler.Scheduler.Scheme.AFTER_PREVIOUS_DONE:
            assert results[2][1] > results[1][1]
            assert results[1][1] > results[0][1]
        elif scheme == scheduler.Scheduler.Scheme.EXECUTION_ORDER:
            assert results[2][1] > results[1][1]
            assert results[1][1] > results[0][1]
        else:
            assert results[0][1] > results[2][1]
            assert results[2][1] > results[1][1]
        sch.shutdown()



"""Module to read and write files in mhealth format
"""


import pandas as pd
from . import helper
from . import constants
from .. import scheduler
import functools
import os
from loguru import logger
import numpy as np
from concurrent import futures


class MhealthFileReader:
    def __init__(self, filepath):
        self._filepath = filepath
        self._data = None
        self._iterator = None

    @staticmethod
    def read_csvs(*filepaths, datetime_cols=[0]):
        dfs = []
        for filepath in filepaths:
            reader = MhealthFileReader(filepath).read_csv(
                datetime_cols=datetime_cols)
            dfs.append(next(reader.get_data()))

        # combine and sort timestamps
        result = pd.concat(dfs, axis=0, ignore_index=True, sort=False)
        result.sort_values(
            by=result.columns[datetime_cols[0]], inplace=True, ignore_index=True)
        return result

    def read_csv(self, chunksize=None, datetime_cols=[0]):
        """
        Known isuee:
        When used with writer, chunksize has to be None
        """
        reader = pd.read_csv(
            self._filepath, parse_dates=datetime_cols, infer_datetime_format=True, chunksize=chunksize, engine='c')
        if type(reader) == pd.DataFrame:
            self._data = reader
        else:
            self._iterator = reader
        return self

    def get_data(self):
        if self._data is not None:
            data = self._data.copy()
            data = helper.format_columns(
                data, filetype=constants.SENSOR_FILE_TYPE)
            yield data
        else:
            for data in self._iterator:
                data = helper.format_columns(
                    data, filetype=constants.SENSOR_FILE_TYPE)
                yield data


class MhealthFileWriter:
    def __init__(self, dataset_path, pid, hourly=False, date_folders=False):
        """
        Known issue: 
        hourly has to be set True
        """
        self._dataset_path = dataset_path
        self._pid = pid
        self._hourly = hourly
        self._date_folders = date_folders
        self._file_type = None
        self._executor = None

    def set_for_sensor(self, sensor_type, data_type, sensor_id, version_code):
        self._sensor_type = sensor_type
        self._data_type = data_type
        self._sensor_id = sensor_id
        self._version_code = version_code
        self._file_type = constants.SENSOR_FILE_TYPE

    def set_for_annotation(self, annotation_type, annotator):
        self._annotation_type = annotation_type
        self._annotator = annotator
        self._file_type = constants.ANNOTATION_FILE_TYPE

    def write_csv(self, data, append=False, block=True):
        writing_tasks = []
        groups = self._split_data(data)
        self._init_executor(len(groups))
        for group in groups:
            task = self._executor.submit(self._write_csv, group, append)
            writing_tasks.append(task)
        if block:
            results = self._executor.get_all_remaining_results()
            if len(results) != len(writing_tasks):
                raise IOError('Some chunks are not written to files correctly')
            else:
                logger.info('All chunks have been written to files.')
                return results
        else:
            return writing_tasks

    def _split_data(self, data):
        # TODO: Annotations should be cropped nicely at the edge of hourly files.
        if self._hourly:
            if self._file_type == constants.SENSOR_FILE_TYPE:
                group_key = constants.TIMESTAMP_COL
            elif self._file_type == constants.ANNOTATION_FILE_TYPE:
                group_key = constants.START_TIME_COL
            groups = data.groupby(pd.Grouper(key=group_key, freq='H'))
            groups = [group for name, group in groups]
            groups = list(filter(lambda group: not group.empty, groups))
        else:
            groups = [data]
        return groups

    def _init_executor(self, n_chunks):
        n = max(min(n_chunks, os.cpu_count() - 8), 1)
        self._executor = scheduler.Scheduler(
            mode=scheduler.Scheduler.Mode.THREAD, scheme=scheduler.Scheduler.Scheme.SUBMIT_ORDER, max_workers=n)

    def _get_output_filename(self, data):
        timestamp_str = helper.format_file_timestamp_from_data(
            data, self._file_type)
        if self._file_type == constants.SENSOR_FILE_TYPE:
            filename = "{}-{}-{}.{}-{}.{}.{}.csv".format(
                self._sensor_type, self._data_type,
                self._version_code, self._sensor_id, self._data_type, timestamp_str, self._file_type)
        else:
            filename = "{}.{}-{}.{}.{}.csv".format(
                self._annotation_type, self._annotator, self._annotation_type, timestamp_str, self._file_type)
        return filename

    def _get_output_folder(self, data):
        base_folder = os.path.join(self._dataset_path, self._pid,
                                   constants.MASTER_FOLDER)
        if self._date_folders:
            date_folder_path = helper.format_date_folder_path_from_data(
                data, self._file_type)
            output_folder = os.path.join(base_folder, date_folder_path)
        else:
            output_folder = base_folder
        return output_folder

    def _get_output_filepath(self, output_folder, output_filename, append):
        if append:
            existing_filenames = os.listdir(output_folder)
            for fname in existing_filenames:
                f1 = os.path.join(output_folder, fname)
                f2 = os.path.join(output_folder, output_filename)
                if helper.compare_two_mhealth_filepaths(
                        f1, f2):
                    return f1
        return os.path.join(output_folder, output_filename)

    def _write_csv(self, data, append):
        output_folder = self._get_output_folder(data)
        os.makedirs(output_folder, exist_ok=True)
        output_filename = self._get_output_filename(data)
        output_filepath = self._get_output_filepath(
            output_folder, output_filename, append)
        text = data.values
        header = ','.join(data.columns.values)

        if self._file_type == constants.SENSOR_FILE_TYPE:
            fmt = ['%.23s'] + ['%.6f'] * (len(data.columns) - 1)
        elif self._file_type == constants.ANNOTATION_FILE_TYPE:
            fmt = ['%.23s', '%.23s', '%.23s', '%s']
        if append == False or not os.path.exists(output_filepath):
            logger.debug(
                'Overwriting existing file for {} if there are any'.format(
                    output_filepath)
            )
            np.savetxt(output_filepath, text, delimiter=",",
                       fmt=fmt, header=header, comments='')
        else:
            logger.debug(
                'Appending to existing file for {} if there are any'.format(output_filepath))
            with open(output_filepath, 'ab') as f:
                np.savetxt(f, text, delimiter=",",
                           fmt=fmt)
        return output_filepath



"""
Datasets for activity recognition. This module provides functions to load the raw, processed datasets. It also provides functions to reproduce processed datasets from raw.

Author: Qu Tang
Date: 01/28/2020
License: GNU v3
"""

import functools
from loguru import logger
import os
import subprocess
import tarfile
import pkg_resources

import numpy as np
import pandas as pd
import wget

from .. import developer, env, generator, moment, segmentor, stream
from .. import mhealth_format as mh
# from . import _process_mhealth, _process_annotations


def get_available_sample_data():
    data_names = os.listdir(pkg_resources.resource_filename(__name__, 'data'))
    data_names = list(map(lambda name: name.replace('.csv', ''), data_names))
    return data_names


def get_sample_datapath(name):
    if name in get_available_sample_data():
        filepath = pkg_resources.resource_filename(
            __name__, f'data/{name}.csv')
    else:
        raise FileNotFoundError('The given sample data name is not supported.')
    return filepath


def get_dataset_names():
    """Report available example datasets, useful for reporting issues."""
    # delayed import to not demand bs4 unless this function is actually used
    return [
        'spades_lab'
    ]


def cache_data(dataset_name, data_home=None):
    if dataset_name == 'spades_lab':
        url = "https://github.com/qutang/MUSS/releases/latest/download/muss_data.tar.gz"
        name = dataset_name + '.tar.gz'
        original_name = 'muss_data'
    dataset_path = os.path.join(env.get_data_home(), dataset_name)
    if os.path.exists(dataset_path):
        return dataset_path
    else:
        compressed_dataset_path = download_dataset(url, name)
        dataset_path = decompress_dataset(
            compressed_dataset_path, original_name)
        os.remove(compressed_dataset_path)
    return dataset_path


def get_dataset_path(dataset_name):
    return cache_data(dataset_name)


def load_dataset(dataset_name):
    dataset_path = cache_data(dataset_name)
    if dataset_name == 'spades_lab':
        return mh.traverse_dataset(dataset_path)


# def process_dataset(dataset_name, approach='muss'):
#     dataset_dict = load_dataset(dataset_name)
#     if dataset_name == 'spades_lab':
#         sr = 80
#         processed_dataset = _process_mhealth.process_mehealth_dataset(
#             dataset_dict, approach=approach, sr=sr)
#     else:
#         raise NotImplementedError('Only "spades_lab" dataset is supported.')
#     output_path = os.path.join(
#         mh.get_processed_path(dataset_dict['meta']['root']), approach + '.csv')
#     os.makedirs(os.path.dirname(output_path), exist_ok=True)
#     processed_dataset.to_csv(
#         output_path, float_format='%.6f', header=True, index=False)
#     logger.info('Processed {} dataset is saved to {}'.format(
#         dataset_name, output_path))
#     dataset_dict['processed'][approach] = output_path
#     return dataset_dict


def download_dataset(url, name):
    spades_lab_url = url
    output_path = os.path.join(env.get_data_home(), name)
    if os.path.exists(output_path):
        return output_path
    else:
        result = wget.download(spades_lab_url, out=output_path)
    return result


def decompress_dataset(dataset_path, original_name):
    cwd = os.path.dirname(dataset_path)
    name = os.path.basename(dataset_path).split('.')[0]
    if developer.command_is_available('tar --version'):
        logger.info('Using system tar command to decompress dataset file')
        decompress_cmd = ['tar', '-xzf', dataset_path]
        subprocess.run(' '.join(decompress_cmd),
                       check=True, shell=True, cwd=cwd)
    else:
        logger.info('Using Python tar module to decompress data file')
        tar = tarfile.open(dataset_path)
        tar.extractall(path=cwd)
        tar.close()
    os.rename(os.path.join(cwd, original_name), os.path.join(cwd, name))
    output_path = os.path.join(cwd, name)
    return output_path


def parse_annotations(dataset_name, annot_df, pid, st, et):
    if dataset_name == 'spades_lab':
        return _process_annotations._parse_spades_lab_annotations(annot_df, pid, st, et)
    else:
        raise NotImplementedError('Only support spades_lab dataset for now')



import pandas as pd
import numpy as np
from .. import segmentor
from .. import moment
import pytest


@pytest.fixture
def sensor_data(spades_lab_data):
    sensor_file = spades_lab_data['subjects']['SPADES_47']['sensors']['DW'][0]
    df = pd.read_csv(sensor_file, header=0, parse_dates=[0])
    return df.iloc[:1000, :]


class TestSegmentor:
    def test_segment(self, sensor_data):
        df = pd.DataFrame(data=sensor_data)
        seg = segmentor.Segmentor()
        i = 0
        for sample, context in seg.segment(df):
            assert len(sample.shape) == 1
            assert len(sample) == 4
            np.testing.assert_array_equal(
                sample.values, sensor_data.iloc[i, :].values)
            i += 1


class TestSlidingWindowSegmentor:
    @pytest.mark.parametrize("window_size", [0, 1, 10])
    @pytest.mark.parametrize('ref_st', [None, pd.Timestamp.now()])
    def test_segment(self, sensor_data, window_size, ref_st):
        if window_size == 0:
            with pytest.raises(ValueError):
                seg = segmentor.SlidingWindowSegmentor(
                    window_size=window_size, ref_st=ref_st, st_col=0, et_col=None)
        else:
            seg = segmentor.SlidingWindowSegmentor(
                window_size=window_size, ref_st=ref_st, st_col=0, et_col=None)
            i = 0
            for sample, context in seg.segment(sensor_data):
                assert moment.Moment.get_duration(
                    context['start_time'], context['stop_time'], unit='s') == window_size
                assert sample.shape[0] <= window_size * 80
                assert sample.shape[1] == 4
                i += 1
            if ref_st is not None:
                assert i == 0



from arus.testing import load_test_data
import pandas as pd
from arus.models.muss import MUSSModel, Strategy
from pathos import pools
import arus
from playsound import playsound
import datetime as dt
import os
import enum
import queue
import numpy as np
from loguru import logger

muss = MUSSModel()


class PROCESSOR_MODE(enum.Enum):
    TEST_ONLY = enum.auto()
    TEST_AND_SAVE = enum.auto()
    COLLECT_ONLY = enum.auto()
    TEST_AND_COLLECT = enum.auto()
    ACTIVE_COLLECT = enum.auto()


def load_origin_dataset():
    class_filepath, _ = load_test_data(file_type='mhealth', sensor_type='class_labels',
                                       file_num='single', exception_type='multi_tasks')
    feature_filepath, _ = load_test_data(file_type='mhealth', sensor_type='feature',
                                         file_num='single', exception_type='multi_placements')
    class_df = pd.read_csv(class_filepath, parse_dates=[
        0, 1, 2], infer_datetime_format=True)
    feature_df = pd.read_csv(feature_filepath, parse_dates=[
        0, 1, 2], infer_datetime_format=True)
    return feature_df, class_df


def get_class_label_candidates(dataset):
    class_df = dataset[1]
    return class_df['MUSS_22_ACTIVITY_ABBRS'].unique().tolist()


def get_model_summary(model=None):
    if model is not None:
        name = ','.join(model[0].classes_)
        acc = round(model[2], 2)
        summary = 'Classes:\n' + name + '\n' + \
            'Training accuracy:\n' + str(acc)
    else:
        summary = 'No model is available'
    return summary


def get_dataset_summary(dataset=None):
    if dataset is not None:
        labels = dataset['GT_LABEL'].unique().tolist()
        num_of_windows = dataset.shape[0]
        summary = 'Classes:\n' + \
            str(labels) + '\n' + 'Total windows:\n' + str(num_of_windows)
    else:
        summary = 'No data is available'
    return summary


def get_data_summary_table(dataset=None):
    def _summarize(df):
        if 'PREDICTION' not in df:
            windows = df.shape[0]
            accuracy = np.nan
        else:
            correct = np.sum(df['PREDICTION'] == df.name)
            incorrect = np.sum(df['PREDICTION'] != df.name)
            accuracy = correct / (correct + incorrect)
            windows = correct + incorrect
        return pd.DataFrame(data={'a': [windows], 'b': [accuracy]})

    if dataset is None:
        return None
    else:
        result = dataset.groupby(by=['GT_LABEL']).apply(
            _summarize).reset_index(drop=False)
        result = result[['GT_LABEL', 'a', 'b']]
        if 'PREDICTION' in dataset:
            accuracy = np.sum(dataset['PREDICTION'] ==
                              dataset['GT_LABEL']) / dataset.shape[0]
        else:
            accuracy = np.nan
        result = result.append(
            {'GT_LABEL': 'Total', 'a': dataset.shape[0], 'b': accuracy}, ignore_index=True)
        return result.values.tolist()


def extract_placement_features(feature_df, placement_names, group_col=[]):
    placement_features = []
    for placement in placement_names:
        placement_feature = feature_df.loc[
            feature_df['SENSOR_PLACEMENT'] == placement,
            ['HEADER_TIME_STAMP', 'START_TIME', 'STOP_TIME'] + group_col +
            muss.get_feature_names()
        ]
        placement_features.append(placement_feature)
    return placement_features


def train_model(origin_labels=None,
                origin_dataset=None,
                origin_model=None,
                new_labels=None,
                new_dataset=None,
                progress_queue=None,
                placement_names=['DW', 'DA', 'DT'], class_col='MUSS_22_ACTIVITY_ABBRS',
                strategy=Strategy.USE_ORIGIN_ONLY,
                pool=None):
    progress_queue = progress_queue or queue.Queue()

    pool = pool or pools.ProcessPool(nodes=1)
    pool.restart(force=True)

    if origin_model is not None:
        placement_names = origin_model[-2]
        origin_labels = origin_labels or origin_model[0].classes_

    if origin_dataset is not None:
        origin_feature, origin_class, origin_feature_names = prepare_origin_dataset(
            origin_dataset,
            origin_labels,
            placement_names,
            class_col,
            progress_queue
        )

    if new_dataset is not None:
        new_feature, new_class, new_feature_names = prepare_new_dataset(
            new_dataset,
            new_labels,
            class_col,
            progress_queue
        )

    progress_queue.put('Select training strategy...')
    if strategy == Strategy.USE_ORIGIN_ONLY:
        input_feature = origin_feature
        input_class = origin_class
        feature_names = origin_feature_names
    elif strategy == Strategy.USE_NEW_ONLY:
        input_feature = new_feature
        input_class = new_class
        feature_names = new_feature_names
    elif strategy == Strategy.REPLACE_ORIGIN:
        input_feature, input_class = replace_original_data(
            origin_feature,
            origin_class,
            origin_labels,
            new_feature,
            new_class,
            new_labels
        )
        feature_names = origin_feature_names
    elif strategy == Strategy.COMBINE_ORIGIN:
        input_feature, input_class = combine_original_data(
            origin_feature,
            origin_class,
            origin_labels,
            new_feature,
            new_class,
            new_labels
        )
        feature_names = origin_feature_names

    progress_queue.put('Training SVM classifier...')
    task = pool.apipe(muss.train_classifier, input_feature,
                      input_class, class_col=class_col, feature_names=feature_names, placement_names=placement_names)
    progress_queue.put(task)


def prepare_origin_dataset(dataset, labels, placement_names, class_col, progress_queue, group_col=[]):
    origin_feature = dataset[0]
    origin_class = dataset[1]

    progress_queue.put('Extracting training data for each placement...')
    placement_features = extract_placement_features(
        origin_feature, placement_names)

    progress_queue.put('Combining training data together...')
    combined_feature_set, combined_feature_names = muss.combine_features(
        *placement_features, placement_names=placement_names)

    cleared_class_set = origin_class[['HEADER_TIME_STAMP',
                                      'START_TIME', 'STOP_TIME'] + group_col + [class_col]]

    progress_queue.put('Synchronizing training data and class labels...')
    synced_feature, synced_class = muss.sync_feature_and_class(
        combined_feature_set, cleared_class_set)

    # only use training labels
    progress_queue.put('Filtering out unused class labels...')
    filter_condition = synced_class[class_col].isin(
        labels)
    input_feature = synced_feature.loc[filter_condition, :]
    input_class = synced_class.loc[filter_condition, :]
    return input_feature, input_class, combined_feature_names


def prepare_new_dataset(dataset, labels, class_col, progress_queue):
    progress_queue.put(
        'Preparing feature and classes for the new collected data...')
    new_feature = dataset.iloc[:, :-2]
    new_class = dataset.iloc[:, [0, 1, 2, -2]]
    new_class = new_class.rename(
        columns={'GT_LABEL': class_col})
    progress_queue.put('Filtering out unused class labels...')
    filter_condition = new_class[class_col].isin(
        labels)
    input_feature = new_feature.loc[filter_condition, :]
    input_class = new_class.loc[filter_condition, :]
    combined_feature_names = new_feature.columns[3:].values.tolist()
    return input_feature, input_class, combined_feature_names


def validate_model(origin_labels=None,
                   origin_dataset=None,
                   origin_model=None,
                   new_labels=None,
                   new_dataset=None,
                   progress_queue=None,
                   placement_names=['DW', 'DA'], class_col='MUSS_22_ACTIVITY_ABBRS',
                   strategy=Strategy.USE_ORIGIN_ONLY,
                   pool=None):
    progress_queue = progress_queue or queue.Queue()

    pool = pool or pools.ProcessPool(nodes=1)
    pool.restart(force=True)
    feature_names = None
    if origin_model is not None:
        placement_names = origin_model[-2]
        origin_labels = origin_labels or origin_model[0].classes_

    if origin_dataset is not None:
        origin_feature, origin_class, origin_feature_names = prepare_origin_dataset(
            origin_dataset,
            origin_labels,
            placement_names,
            class_col,
            progress_queue,
            group_col=['PID']
        )
        feature_names = origin_feature_names

    new_feature = None
    new_class = None
    if new_dataset is not None:
        new_feature, new_class, new_feature_names = prepare_new_dataset(
            new_dataset,
            new_labels,
            class_col,
            progress_queue
        )
        if feature_names is None:
            feature_names = new_feature_names
        new_feature = new_feature[
            ['HEADER_TIME_STAMP', 'START_TIME', 'STOP_TIME'] + feature_names]

    progress_queue.put('Validating SVM classifier...')

    task = pool.apipe(muss.validate_classifier, origin_feature,
                      origin_class, class_col=class_col, feature_names=feature_names, placement_names=placement_names, new_input_feature=new_feature, new_input_class=new_class, strategy=strategy, group_col='PID')
    progress_queue.put(task)


def replace_original_data(origin_feature, origin_class, origin_labels, new_feature, new_class, new_labels):
    labels_keep_in_origin = list(filter(
        lambda l: l not in new_labels, origin_labels))
    filter_condition = origin_class['MUSS_22_ACTIVITY_ABBRS'].isin(
        labels_keep_in_origin)
    origin_feature = origin_feature.loc[filter_condition, :]
    origin_class = origin_class.loc[filter_condition, :]

    input_feature = pd.concat(
        [origin_feature, new_feature], sort=False, join='outer')
    input_class = pd.concat([origin_class, new_class],
                            sort=False, join='outer')
    return input_feature, input_class


def combine_original_data(origin_feature, origin_class, origin_labels, new_feature, new_class, new_labels):
    labels_keep_in_origin = origin_labels
    filter_condition = origin_class['MUSS_22_ACTIVITY_ABBRS'].isin(
        labels_keep_in_origin)
    origin_feature = origin_feature.loc[filter_condition, :]
    origin_class = origin_class.loc[filter_condition, :]

    new_filter_condition = new_class['MUSS_22_ACTIVITY_ABBRS'].isin(new_labels)
    new_feature = new_feature.loc[new_filter_condition, :]
    new_class = new_class.loc[new_filter_condition, :]

    input_feature = pd.concat(
        [origin_feature, new_feature], sort=False, join='outer')
    input_class = pd.concat([origin_class, new_class],
                            sort=False, join='outer')
    return input_feature, input_class


def merge_features(feature_set, placement_names, muss, group_col=None):
    placement_features = []
    if group_col is None:
        selected_cols = [
            'HEADER_TIME_STAMP', 'START_TIME', 'STOP_TIME'] + muss.get_feature_names()
    else:
        selected_cols = [
            'HEADER_TIME_STAMP', 'START_TIME', 'STOP_TIME', group_col] + muss.get_feature_names()
    for placement in placement_names:
        place_features = feature_set.loc[feature_set['SENSOR_PLACEMENT']
                                         == placement, selected_cols]
        placement_features.append(place_features)
    combined_feature_set, combined_feature_names = muss.combine_features(
        *placement_features, placement_names=placement_names, group_col=group_col)
    return combined_feature_set, combined_feature_names


def get_confusion_matrix_figure(validation_result, fig=None):
    muss = MUSSModel()
    labels = validation_result[-2]
    fig = muss.get_confusion_matrix(
        validation_result[0], validation_result[1], labels=labels, graph=True, fig=fig)
    return fig


def get_classification_report_table(validation_result):
    muss = MUSSModel()
    report = muss.get_classification_report(
        validation_result[0], validation_result[1], labels=validation_result[-2])
    report_table = []
    for key, values in report.items():
        report_row = []
        if key in validation_result[-2]:
            report_row.append(key)
            for metric_name, metric_value in values.items():
                if metric_name == 'precision':
                    report_row.append(round(metric_value, 2))
                elif metric_name == 'recall':
                    report_row.append(round(metric_value, 2))
                elif metric_name == 'f1-score':
                    report_row.append(round(metric_value, 2))
            report_table.append(report_row)
    return report_table


def connect_devices(devices, model, placement_names=['DW', 'DA'], mode=PROCESSOR_MODE.TEST_ONLY, output_folder=None, pid=None, pool=None):
    logger.info('device addrs: ' + str(devices))
    logger.info('device placements: ' + str(placement_names))
    pool = pool or pools.ThreadPool(nodes=1)
    pool.restart(force=True)
    device_addrs = devices
    streams = []
    start_time = dt.datetime.now()
    output_folder = os.path.join(output_folder, 'data')
    if model is not None:
        placement_names = model[-2]
    kwargs = {}
    for addr, placement in zip(device_addrs, placement_names):
        generator = arus.plugins.metawear.MetaWearAccelDataGenerator(
            addr, sr=50, grange=8, buffer_size=100)
        segmentor = arus.segmentor.SlidingWindowSegmentor(window_size=4)
        stream = arus.Stream(generator, segmentor, name=placement)
        streams.append(stream)
        kwargs[placement] = {'sr': 50}
    if mode == PROCESSOR_MODE.TEST_ONLY:
        pipeline = muss.get_inference_pipeline(
            *streams, name='muss-pipeline', model=model, max_processes=2, scheduler='processes', **kwargs)
    else:
        pipeline = muss.get_data_collection_pipeline(
            *streams, name='muss-pipeline', model=model, max_processes=2, scheduler='processes', output_folder=output_folder, pid=pid, **kwargs
        )
    task = pool.apipe(pipeline.connect, start_time=start_time)
    return task


def disconnect_devices(pipeline, pool=None):
    pool = pool or pools.ThreadPool(nodes=1)
    pool.restart(force=True)
    task = pool.apipe(pipeline.stop)
    return task


def start_test_model(pipeline, pool=None):
    start_time = dt.datetime.now()
    pool = pool or pools.ThreadPool(nodes=1)
    pool.restart(force=True)
    task = pool.apipe(pipeline.process, start_time=start_time)
    return task


def stop_test_model(pipeline, pool=None):
    pool = pool or pools.ThreadPool(nodes=1)
    pool.restart(force=True)
    task = pool.apipe(pipeline.pause)
    return task


def save_annotation(current_annotation, output_folder, pid, session_name, pool=None):
    output_folder = os.path.join(output_folder, 'data')
    writer = arus.mh.MhealthFileWriter(
        output_folder, pid, hourly=False, date_folders=False)
    writer.set_for_annotation(session_name, 'ARUS')
    df = pd.DataFrame.from_dict(current_annotation)
    task = writer.write_csv(df, append=True, block=False)[0]
    return task


def play_sound(text, pool=None):
    pool = pool or pools.ThreadPool(nodes=1)
    pool.restart(force=True)
    return pool.apipe(playsound, text + '.mp3', block=True)


def get_nearby_devices(pool=None):
    pool = pool or pools.ThreadPool(nodes=1)
    pool.restart(force=True)
    scanner = arus.plugins.metawear.MetaWearScanner()
    task = pool.apipe(scanner.get_nearby_devices, max_devices=3)
    return task



"""
Computing features of descriptive statistics for accelerometer data
"""
import numpy as np
from scipy import stats as sp_stats
from .. import extensions as ext


STAT_FEATURE_NAME_PREFIX = [
    'MEAN',
    'MEDIAN',
    'STD',
    'SKEW',
    'KURTOSIS',
    'MAX',
    'MIN',
    'RANGE',
    'ABS_MAX',
    'ABS_MIN',
    'ZCR',
    'MCR',
    'TOTAL_POWER',
    'CORRELATION'
]


def mean(X):
    X = ext.numpy.atleast_float_2d(X)
    result = np.nanmean(X, axis=0, keepdims=True)
    return result, [f'{STAT_FEATURE_NAME_PREFIX[0]}_{i}' for i in range(X.shape[1])]


def median(X):
    X = ext.numpy.atleast_float_2d(X)
    result = np.nanmedian(X, axis=0, keepdims=True)
    return result, [f'{STAT_FEATURE_NAME_PREFIX[1]}_{i}' for i in range(X.shape[1])]


def std(X):
    X = ext.numpy.atleast_float_2d(X)
    result = np.nanstd(X, axis=0, ddof=1, keepdims=True)
    return result, [f'{STAT_FEATURE_NAME_PREFIX[2]}_{i}' for i in range(X.shape[1])]


def skew(X):
    X = ext.numpy.atleast_float_2d(X)
    result = np.atleast_2d(sp_stats.skew(X, axis=0, nan_policy='omit'))
    result = result, [
        f'{STAT_FEATURE_NAME_PREFIX[3]}_{i}' for i in range(X.shape[1])]
    return result


def kurtosis(X):
    X = ext.numpy.atleast_float_2d(X)
    result = np.atleast_2d(sp_stats.kurtosis(X, axis=0, nan_policy='omit'))
    result = result, [
        f'{STAT_FEATURE_NAME_PREFIX[4]}_{i}' for i in range(X.shape[1])]
    return result


def max_value(X):
    X = ext.numpy.atleast_float_2d(X)
    result = np.nanmax(X, axis=0, keepdims=True)
    return result, [f'{STAT_FEATURE_NAME_PREFIX[5]}_{i}' for i in range(X.shape[1])]


def min_value(X):
    X = ext.numpy.atleast_float_2d(X)
    result = np.nanmin(X, axis=0, keepdims=True)
    return result, [f'{STAT_FEATURE_NAME_PREFIX[6]}_{i}' for i in range(X.shape[1])]


def max_minus_min(X):
    X = ext.numpy.atleast_float_2d(X)
    result = max_value(X)[0] - min_value(X)[0]
    return result, [f'{STAT_FEATURE_NAME_PREFIX[7]}_{i}' for i in range(X.shape[1])]


def abs_max_value(X):
    X = ext.numpy.atleast_float_2d(X)
    result = np.nanmax(np.abs(X), axis=0, keepdims=True)
    return result, [f'{STAT_FEATURE_NAME_PREFIX[8]}_{i}' for i in range(X.shape[1])]


def abs_min_value(X):
    X = ext.numpy.atleast_float_2d(X)
    result = np.nanmin(np.abs(X), axis=0, keepdims=True)
    return result, [f'{STAT_FEATURE_NAME_PREFIX[9]}_{i}' for i in range(X.shape[1])]


def zcr(X):
    """Compute zero crossing rate
    """
    X = ext.numpy.atleast_float_2d(X)
    X_offset = X[1:, :]
    X_orig = X[:-1, :]
    zero_crossings = np.sum(np.multiply(X_orig, X_offset)
                            < 0, axis=0, keepdims=True)
    zc_rates = zero_crossings / (X.shape[0] - 1)
    return zc_rates, [f'{STAT_FEATURE_NAME_PREFIX[10]}_{i}' for i in range(X.shape[1])]


def mcr(X):
    """Compute mean crossing rate
    """
    X = ext.numpy.atleast_float_2d(X)
    X_demean = X - np.mean(X, axis=0, keepdims=True)
    mc_rates, _ = zcr(X_demean)
    return mc_rates, [f'{STAT_FEATURE_NAME_PREFIX[11]}_{i}' for i in range(X.shape[1])]


def total_power(X):
    """Compute total power
    """
    X = ext.numpy.atleast_float_2d(X)
    X = np.power(X, 2)
    total_energy = np.sum(X, axis=0, keepdims=True)
    result = total_energy / X.shape[0]
    return result, [f'{STAT_FEATURE_NAME_PREFIX[12]}_{i}' for i in range(X.shape[1])]


def correlation(X):
    X = ext.numpy.atleast_float_2d(X)
    if X.shape[1] != 3:
        result = np.repeat(np.nan, X.shape[1])
    else:
        corr_mat = np.corrcoef(X, rowvar=False)
        if np.isscalar(corr_mat) and np.isnan(corr_mat):
            result = np.repeat(np.nan, X.shape[1])
        else:
            inds = np.tril_indices(
                n=corr_mat.shape[0], k=-1, m=corr_mat.shape[1])
            result = []
            for i, j in zip(inds[0], inds[1]):
                result.append(corr_mat[i, j])
    result = np.atleast_2d(result)
    return result, [f'{STAT_FEATURE_NAME_PREFIX[13]}_{i}' for i in range(result.shape[1])]


def stat_features(X, selected=STAT_FEATURE_NAME_PREFIX):
    X = ext.numpy.atleast_float_2d(X)
    stat_funcs = [
        mean,
        median,
        std,
        skew,
        kurtosis,
        max_value,
        min_value,
        max_minus_min,
        abs_max_value,
        abs_min_value,
        zcr,
        mcr,
        total_power,
        correlation
    ]

    fv = []
    fv_names = []
    for func, prefix in zip(stat_funcs, STAT_FEATURE_NAME_PREFIX):
        if prefix in selected:
            result, names = func(X)
            fv.append(result)
            fv_names += names

    if len(fv) == 0:
        return None, None

    result = np.concatenate(fv, axis=1)
    return result, fv_names



import numpy as np
import pandas as pd
from . import stream2, generator, segmentor, synchronizer, processor, scheduler, pipeline, node
from loguru import logger
import sys
from . import extensions as ext
from . import mhealth_format as mh
from .error_code import ErrorCode
import tqdm


class ClassSet:
    def __init__(self, raw_sources, aids):
        self._raw_sources = raw_sources
        self._aids = aids

    def _validate_input_as_df(self):
        if len(self._raw_sources) == 0:
            return
        if type(self._raw_sources[0]) is not pd.DataFrame:
            logger.error(
                '[Error code: {ErrorCode.INPUT_ARGUMENT_FORMAT_ERROR.name}] To compute class labels offline, the input raw data should be annotation dataframe stored in mhealth format.')
            sys.exit(ErrorCode.INPUT_ARGUMENT_FORMAT_ERROR.name)

    def _validate_input_as_generator(self):
        if type(self._raw_sources[0]) is not generator.Generator:
            logger.error(
                f'[Error code: {ErrorCode.INPUT_ARGUMENT_FORMAT_ERROR.name}] To compute class labels online, the input raw data should be arus Generator object.')
            sys.exit(ErrorCode.INPUT_ARGUMENT_FORMAT_ERROR.name)

    def compute_offline(self, window_size, class_func, task_names, start_time=None, stop_time=None, step_size=None, show_progress=True, **kwargs):
        self._validate_input_as_df()
        self._task_names = task_names
        step_size = step_size or window_size
        window_start_markers = ext.pandas.split_into_windows(
            *self._raw_sources, step_size=step_size, st=start_time, et=stop_time, st_col=1, et_col=2)
        class_vectors = []
        with tqdm.tqdm(total=len(window_start_markers), disable=not show_progress) as bar:
            for window_st in window_start_markers:
                window_et = window_st + pd.Timedelta(window_size, unit='s')
                dfs = []
                for raw_df in self._raw_sources:
                    df = ext.pandas.segment_by_time(
                        raw_df, seg_st=window_st, seg_et=window_et, st_col=1, et_col=2)
                    dfs.append(df)
                class_vector = class_func(*dfs, st=window_st, et=window_et,
                                          task_names=task_names, aids=self._aids, **kwargs)
                class_vectors.append(class_vector)
                bar.update()
                bar.set_description(
                    f'Computed class set for window: {window_st}')

        if len(class_vectors) == 0:
            self._class_set = None
        else:
            self._class_set = pd.concat(
                class_vectors, axis=0, ignore_index=True, sort=False)

    def compute_per_window(self, class_func, task_names, **kwargs):
        self._validate_input_as_df()
        class_vector = class_func(
            self._raw_dfs, task_names=task_names, aids=self._aids, **kwargs)
        self._class_set = class_vector
        self._task_names = task_names

    def compute_online(self, seg, class_func, task_names, start_time, aids, **kwargs):
        src_streams = []
        for src_generator, aid in zip(self._raw_sources, self._aids):
            src_stream = stream2.Stream(
                src_generator, seg, name=f'{aid}-stream')
            src_stream.set_essential_context(
                start_time=start_time, stream_id=aid)
            src_streams.append(src_stream)

        sync = synchronizer.Synchronizer()
        sync.add_sources(n=len(self._raw_sources))

        proc = processor.Processor(class_func,
                                   mode=scheduler.Scheduler.Mode.PROCESS,
                                   scheme=scheduler.Scheduler.Scheme.SUBMIT_ORDER,
                                   max_workers=10)
        proc.set_context(**kwargs)

        self._pip = node.Node(op=pipeline.Pipeline(*src_streams,
                                                   synchronizer=sync,
                                                   processor=proc, name='online-classset-pipeline'),
                              t=node.Node.Type.INPUT, name='online-classet-pipeline')
        self._pip.start()
        for pack in self._pip.produce():
            if pack.signal == node.Node.Signal.DATA:
                if pack.values is not None:
                    yield pack.values
            elif pack.signal == node.Node.Signal.STOP:
                break
        self._pip.stop()

    def reset(self):
        self._class_set = None
        self._task_names = None

    def get_class_set(self):
        if self._class_set is not None:
            return self._class_set.copy(deep=True)
        else:
            return None

    def get_task_names(self):
        return self._task_names

    @staticmethod
    def get_annotation_durations(annot_df, label_col):
        durations = annot_df.groupby(label_col).apply(
            lambda rows: np.sum(rows.loc[:, mh.STOP_TIME_COL] - rows.loc[:, mh.START_TIME_COL]))
        return durations
